import os
import subprocess
import sys
import unittest
import warnings

import pandas as pd
import rootpath
import torch
import yaml

from astir.astir import Astir

# import os
# import sys
# import rootpath
#
# module_path = rootpath.detect()
# if module_path not in sys.path:
#     sys.path.append(module_path)
# print(sys.path)
#
# module_path = os.path.join(rootpath.detect(), "astir")
# if module_path not in sys.path:
#     sys.path.append(module_path)
# print(sys.path)


class TestBinAstir(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestBinAstir, self).__init__(*args, **kwargs)
        self.exec_path = os.path.join(rootpath.detect(), "bin/astir")
        self.expr_csv_file = os.path.join(
            os.path.dirname(__file__), "test-data/test_data.csv"
        )
        self.marker_yaml_file = os.path.join(
            os.path.dirname(__file__), "test-data/jackson-2020-markers.yml"
        )
        self.output_file = os.path.join(os.path.dirname(__file__), "output")

        self.expr = pd.read_csv(self.expr_csv_file, index_col=0)
        with open(self.marker_yaml_file, "r") as stream:
            self.marker_dict = yaml.safe_load(stream)

    def test_basic_command(self):
        warnings.filterwarnings("ignore", category=UserWarning)

        bash_command = "python -W ignore {} {} {} {} {}".format(
            self.exec_path,
            "state",
            self.expr_csv_file,
            self.marker_yaml_file,
            self.output_file,
        )
        process = subprocess.Popen(bash_command.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        self.assertIsNone(error)

        read_output = pd.read_csv(self.output_file, index_col=0)
        self.assertEqual(len(read_output), len(self.expr))

        states = self.marker_dict["cell_states"].keys()
        self.assertEqual(len(read_output.columns), len(states))

    def test_command_all_flags(self):
        warnings.filterwarnings("ignore", category=UserWarning)
        (
            design,
            max_epochs,
            lr,
            batch_size,
            random_seed,
            n_init,
            n_init_epochs,
            dtype,
            delta_loss,
            delta_loss_batch,
        ) = (None, 2, 1e-1, 128, 1234, 1, 1, torch.float64, 1e-3, 10)
        bash_command = "python -W ignore {} {} {} {} {}".format(
            self.exec_path,
            "state",
            self.expr_csv_file,
            self.marker_yaml_file,
            self.output_file,
        )
        bash_command += " --design {}".format(design)
        bash_command += " --max_epochs {}".format(max_epochs)
        bash_command += " --learning_rate {}".format(lr)
        bash_command += " --batch_size {}".format(batch_size)
        bash_command += " --random_seed {}".format(random_seed)
        bash_command += " --n_init {}".format(n_init)
        bash_command += " --n_init_epochs {}".format(n_init_epochs)
        bash_command += " --dtype {}".format(dtype)
        bash_command += " --delta_loss {}".format(delta_loss)
        bash_command += " --delta_loss_batch {}".format(delta_loss_batch)
        process = subprocess.Popen(bash_command.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        self.assertIsNone(error)

        # Create Astir object to compare
        ast = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=design,
            random_seed=random_seed,
            dtype=dtype,
        )

        ast.fit_state(
            max_epochs=max_epochs,
            learning_rate=lr,
            batch_size=batch_size,
            delta_loss=delta_loss,
            n_init=n_init,
            n_init_epochs=n_init_epochs,
            delta_loss_batch=delta_loss_batch,
        )

        expected_assign = ast.get_cellstates()
        actual_assign = pd.read_csv(self.output_file, index_col=0)
        self.assertEqual(len(expected_assign), len(actual_assign))
        self.assertTrue((expected_assign.columns == actual_assign.columns).all())

        self.assertTrue(
            (abs(actual_assign.to_numpy() - expected_assign.to_numpy()) < 0.01).all()
        )



import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch

from astir.data import SCDataset


class AstirModel:
    """Abstract class to perform statistical inference to assign. This module is the super class of
    `CellTypeModel` and `CellStateModel` and is not supposed to be instantiated.
    """

    def __init__(
        self,
        dset: Optional[SCDataset],
        random_seed: int,
        dtype: torch.dtype,
        device: torch.device = torch.device("cpu"),
    ) -> None:

        if not isinstance(random_seed, int):
            raise NotClassifiableError("Random seed is expected to be an integer.")
        torch.manual_seed(random_seed)
        np.random.seed(random_seed)

        if dtype != torch.float32 and dtype != torch.float64:
            raise NotClassifiableError(
                "dtype must be one of torch.float32 and torch.float64."
            )
        elif dset is not None and dtype != dset.get_dtype():
            raise NotClassifiableError("dtype must be the same as `dset`.")
        self._dtype: torch.dtype = dtype
        self._data: Dict[str, torch.Tensor] = {}
        self._variables: Dict[str, torch.Tensor] = {}
        self._losses: torch.Tensor = torch.tensor([], dtype=self._dtype)
        self._assignment: pd.DataFrame = pd.DataFrame()

        self._dset = dset
        # self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self._device = device
        self._is_converged = False

    def get_losses(self) -> torch.Tensor:
        """Getter for losses.

        :return: self.losses
        """
        if len(self._losses) == 0:
            raise Exception("The model has not been trained yet")
        return self._losses

    def get_scdataset(self) -> SCDataset:
        """Getter for the `SCDataset`.

        :return: `self._dset`
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        return self._dset

    def get_data(self) -> Dict[str, torch.Tensor]:
        """Get model data

        :return: data
        """
        if self._data == {}:
            raise Exception("The model has not been initialized yet")
        return self._data

    def get_variables(self) -> Dict[str, torch.Tensor]:
        """Returns all variables

        :return: self._variables
        """
        if self._variables == {}:
            raise Exception("The model has not been initialized yet")
        return self._variables

    def is_converged(self) -> bool:
        """Returns True if the model converged

        :return: self._is_converged
        """
        return self._is_converged

    def get_assignment(self) -> pd.DataFrame:
        """Get the final assignment of the dataset.

        :return: the final assignment of the dataset
        """
        if self._assignment.shape == (0, 0):
            raise Exception("The model has not been trained yet")
        return self._assignment

    def _param_init(self) -> None:
        """Initializes parameters and design matrices."""
        raise NotImplementedError("AbstractModel is not supposed to be instantiated.")

    def _forward(
        self, Y: torch.Tensor, X: torch.Tensor, design: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        """One forward pass"""
        raise NotImplementedError("AbstractModel is not supposed to be instantiated.")

    def fit(
        self,
        max_epochs: int,
        learning_rate: float,
        batch_size: int,
        delta_loss: float,
        delta_loss_batch: int,
        msg: str,
    ) -> None:
        """Runs train loops until the convergence reaches delta_loss for
        delta_loss_batch sizes or for max_epochs number of times
        """
        raise NotImplementedError("AbstractModel is not supposed to be instantiated.")


class NotClassifiableError(RuntimeError):
    """Raised when the input data is not classifiable."""

    pass



import contextlib
import os
import subprocess as sp
import warnings
from unittest import TestCase

import h5py
import numpy as np
import pandas as pd
import torch
import yaml

from astir import Astir
from astir.data import SCDataset, from_anndata_yaml, from_csv_dir_yaml, from_csv_yaml


class TestAstir(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestAstir, self).__init__(*args, **kwargs)
        warnings.filterwarnings("ignore", category=UserWarning)
        self.expr_csv_file = os.path.join(
            os.path.dirname(__file__), "test-data/test_data.csv"
        )
        self.marker_yaml_file = os.path.join(
            os.path.dirname(__file__), "test-data/jackson-2020-markers.yml"
        )
        self.design_file = os.path.join(
            os.path.dirname(__file__), "test-data/design.csv"
        )
        self.test_dir = os.path.join(
            os.path.dirname(__file__), "test-data/test-dir-read"
        )
        self.adata_file = os.path.join(
            os.path.dirname(__file__), "test-data/adata_small.h5ad"
        )

        self.expr = pd.read_csv(self.expr_csv_file, index_col=0)
        with open(self.marker_yaml_file, "r") as stream:
            self.marker_dict = yaml.safe_load(stream)

        self.a = Astir(self.expr, self.marker_dict)
        self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def test_basic_instance_creation(self):
        """Tests basic instance creation"""

        self.assertIsInstance(self.a, Astir)
        # self.assertTrue(isinstance(a, str))

    def test_csv_reading(self):
        """Test from_csv_yaml function"""
        a = from_csv_yaml(self.expr_csv_file, self.marker_yaml_file)

        self.assertIsInstance(a, Astir)

    def test_dir_reading(self):

        a = from_csv_dir_yaml(self.test_dir, self.marker_yaml_file)

        self.assertIsInstance(a, Astir)

        ## Make sure the design matrix has been constructed correctly
        self.assertTrue(a._type_dset._design.shape[0] == len(a._type_dset))
        files = os.listdir(self.test_dir)
        files = [f for f in files if f.endswith(".csv")]
        self.assertTrue(a._type_dset._design.shape[1] == len(files))

    def test_csv_reading_with_design(self):

        a = from_csv_yaml(
            self.expr_csv_file, self.marker_yaml_file, design_csv=self.design_file
        )

        self.assertIsInstance(a, Astir)

    def test_fitting_type(self):

        epochs = 2
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                self.a.fit_type(max_epochs=epochs)

        # Check probability matrix looks ok
        probabilities = self.a.get_celltype_probabilities()

        self.assertTrue(probabilities.shape[0] == self.expr.shape[0])
        self.assertIsInstance(probabilities, pd.DataFrame)

        # Check assignments look ok
        assignments = self.a.get_celltypes()
        self.assertIsInstance(assignments, pd.DataFrame)
        self.assertTrue(assignments.shape[0] == self.expr.shape[0])
        self.assertTrue(assignments.columns[0] == "cell_type")

        # Check diagnostics look ok
        type_diagnostics = self.a.diagnostics_celltype(threshold=0.2, alpha=0)
        self.assertIsInstance(type_diagnostics, pd.DataFrame)
        self.assertTrue(
            type_diagnostics.shape[1] == 7
        )  # make sure we have the standard 6 columns

    def test_no_overlap(self):
        bad_file = os.path.join(os.path.dirname(__file__), "test-data/bad_data.csv")
        bad_data = pd.read_csv(bad_file)
        raised = False
        try:
            test = Astir(bad_data, self.marker_dict)
        except (RuntimeError):
            raised = True
        self.assertTrue(raised == True)

    def test_missing_marker(self):
        bad_marker = os.path.join(os.path.dirname(__file__), "test-data/bad_marker.yml")
        with open(bad_marker, "r") as stream:
            bad_dict = yaml.safe_load(stream)
        raised = False
        try:
            test = Astir(self.expr, bad_dict)
            test.fit_type()
        except (RuntimeError):
            raised = True
        self.assertTrue(raised)

    # # Uncomment below test functions to test private variables
    # # Commented it out because these tests can be highly overlapping with
    # # future unittests

    def test_sanitize_dict_state(self):
        """Testing the method _sanitize_dict"""
        expected_state_dict = self.marker_dict["cell_states"]
        (_, actual_state_dict, _) = self.a._sanitize_dict(self.marker_dict)

        expected_state_dict = {
            i: sorted(j) if isinstance(j, list) else j
            for i, j in expected_state_dict.items()
        }
        actual_state_dict = {
            i: sorted(j) if isinstance(j, list) else j
            for i, j in actual_state_dict.items()
        }

        self.assertDictEqual(
            expected_state_dict,
            actual_state_dict,
            "state_dict is different from its expected value",
        )

    def test_state_names(self):
        """Test _state_names field"""
        expected_state_names = sorted(self.marker_dict["cell_states"].keys())
        (_, actual_state_dict, _) = self.a._sanitize_dict(self.marker_dict)
        actual_state_names = sorted(actual_state_dict.keys())

        self.assertListEqual(
            expected_state_names, actual_state_names, "unexpected state_names value"
        )

    def test_celltype_same_seed_same_result(self):
        """Test whether the loss after one epoch one two different models
        with the same random seed have the same losses after one epochs
        """
        warnings.filterwarnings("ignore", category=UserWarning)
        model1 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=42,
        )
        model2 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=42,
        )

        model1.fit_type(max_epochs=10)
        model1_loss = model1.get_type_losses()
        model2.fit_type(max_epochs=10)
        model2_loss = model2.get_type_losses()

        self.assertTrue(np.abs(model1_loss - model2_loss)[-1] < 1e-6)

    def test_celltype_diff_seed_diff_result(self):
        """Test whether the loss after one epoch one two different models
        with the different random seed have different losses after one epoch
        """
        warnings.filterwarnings("ignore", category=UserWarning)
        model1 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=42,
        )
        model2 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=1234,
        )

        model1.fit_type(max_epochs=10)
        model1_loss = model1.get_type_losses()
        model2.fit_type(max_epochs=10)
        model2_loss = model2.get_type_losses()

        self.assertFalse(np.abs(model1_loss - model2_loss)[-1] < 1e-6)

    def test_cellstate_same_seed_same_result(self):
        """Test whether the loss after one epoch one two different models
        with the same random seed have the same losses after one epochs
        """
        warnings.filterwarnings("ignore", category=UserWarning)
        model1 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=42,
        )
        model2 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=42,
        )

        model1.fit_state(max_epochs=5)
        model1_loss = model1.get_state_losses()
        model2.fit_state(max_epochs=5)
        model2_loss = model2.get_state_losses()

        self.assertTrue(np.abs(model1_loss - model2_loss)[-1] < 1e-6)

    # @pytest.mark.filterwarnings("ignore")
    def test_cellstate_diff_seed_diff_result(self):
        """Test whether the loss after one epoch one two different models
        with the different random seed have different losses after one epoch
        """
        warnings.filterwarnings("ignore", category=UserWarning)
        model1 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=42,
        )
        model2 = Astir(
            input_expr=self.expr,
            marker_dict=self.marker_dict,
            design=None,
            random_seed=1234,
        )

        model1.fit_state(max_epochs=5)
        model1_loss = model1.get_state_losses()
        model2.fit_state(max_epochs=5)
        model2_loss = model2.get_state_losses()

        self.assertFalse(np.abs(model1_loss - model2_loss)[-1] < 1e-6)

    def test_cellstate_assignment(self):
        warnings.filterwarnings("ignore", category=UserWarning)
        self.a.fit_state(max_epochs=50, n_init=1)

        state_assignments = self.a.get_cellstates()

        n_classes = len(list(self.marker_dict.keys()))

        self.assertTrue(state_assignments.shape, (len(self.expr), n_classes))

    def test_cellstate_predicted_assignment(self):
        warnings.filterwarnings("ignore", category=UserWarning)
        dset = SCDataset(
            expr_input=self.expr,
            marker_dict=self.marker_dict["cell_states"],
            design=None,
            include_other_column=False,
            device=self._device,
        )

        self.a.fit_state(max_epochs=50, n_init=1)

        state_assignments = self.a.predict_cellstates(dset)

        self.assertTrue(state_assignments.shape, (len(dset), dset.get_n_classes()))

    def test_celltype_assignment(self):
        warnings.filterwarnings("ignore", category=UserWarning)
        self.a.fit_type(max_epochs=50, n_init=1)

        type_assignments = self.a.get_celltypes()

        n_classes = len(list(self.marker_dict.keys()))

        self.assertTrue(type_assignments.shape, (len(self.expr), n_classes + 1))

    def test_celltype_predicted_assignment(self):
        warnings.filterwarnings("ignore", category=UserWarning)

        self.a.fit_type(max_epochs=50, n_init=1)

        type_predict = self.a.predict_celltypes()
        type_assignment = self.a.get_celltype_probabilities()
        comp = type_predict == type_assignment
        self.assertTrue(comp.all().all())

    # def test_adata_reading(self):
    #     ast = from_anndata_yaml(
    #         self.adata_file,
    #         self.marker_yaml_file,
    #         protein_name="protein",
    #         cell_name="cell_name",
    #         batch_name="batch",
    #     )

    #     self.assertTrue(ast.get_type_dataset().get_n_features() == 14)
    #     self.assertTrue(ast.get_type_dataset().get_n_classes() == 6)
    #     self.assertIsInstance(ast.get_type_dataset().get_exprs(), torch.Tensor)
    #     self.assertEqual(ast.get_type_dataset().get_exprs().shape[0], 10)

    def test_cellstate_diagnostics(self):
        warnings.filterwarnings("ignore", category=UserWarning)
        self.a.fit_state(max_epochs=50, n_init=1)

        state_diagnostics = self.a.diagnostics_cellstate()
        self.assertIsInstance(state_diagnostics, pd.DataFrame)

    def test_type_hdf5_summary(self):
        hdf5_summary = "celltype_summary.hdf5"
        info = {
            "max_epochs": 5,
            "learning_rate": 0.001,
            "batch_size": 24,
            "delta_loss": 0.001,
            "n_init": 1,
            "n_init_epochs": 1,
        }
        self.a.fit_type(
            max_epochs=info["max_epochs"],
            learning_rate=info["learning_rate"],
            batch_size=info["batch_size"],
            delta_loss=info["delta_loss"],
            n_init=info["n_init"],
            n_init_epochs=info["n_init_epochs"],
        )
        self.a.save_models(hdf5_summary)
        params = list(self.a.get_type_model().get_data().items()) + list(
            self.a.get_type_model().get_variables().items()
        )

        recog_params = []
        for key, val in self.a.get_type_model().get_recognet().named_parameters():
            recog_params.append((key, val.detach().cpu().numpy()))
        same = True
        with h5py.File(hdf5_summary, "r") as f:
            f_params = f["/celltype_model/parameters"]
            for key, val in params:
                if not (val.detach().cpu().numpy() == f_params[key][()]).all().all():
                    same = False
            f_recog = f["/celltype_model/recog_net"]
            for key, val in recog_params:
                if not (f_recog[key][()] == val).all():
                    same = False
            f_info = f["/celltype_model/run_info"]
            for key, val in info.items():
                if val != f_info[key][()]:
                    same = False
            if not (
                self.a.get_type_model().get_losses().cpu().numpy()
                == f["/celltype_model/losses"]["losses"][()]
            ).all():
                same = False
        self.assertTrue(same)

    def test_state_summary(self):
        hdf5_summary = "cellstate_summary.hdf5"
        info = {
            "max_epochs": 5,
            "learning_rate": 0.001,
            "batch_size": 24,
            "delta_loss": 0.001,
            "n_init": 1,
            "n_init_epochs": 1,
            "delta_loss_batch": 2,
        }
        self.a.fit_state(
            max_epochs=info["max_epochs"],
            learning_rate=info["learning_rate"],
            batch_size=info["batch_size"],
            delta_loss=info["delta_loss"],
            n_init=info["n_init"],
            n_init_epochs=info["n_init_epochs"],
            delta_loss_batch=info["delta_loss_batch"],
        )
        self.a.save_models(hdf5_summary)
        params = list(self.a.get_state_model().get_data().items()) + list(
            self.a.get_state_model().get_variables().items()
        )
        recog_params = []
        for key, val in self.a.get_state_model().get_recognet().named_parameters():
            recog_params.append((key, val.detach().cpu().numpy()))
        same = True
        with h5py.File(hdf5_summary, "r") as f:
            f_params = f["/cellstate_model/parameters"]
            for key, val in params:
                if not (val.detach().cpu().numpy() == f_params[key][()]).all().all():
                    same = False
            f_recog = f["/cellstate_model/recog_net"]
            for key, val in recog_params:
                if not (f_recog[key][()] == val).all():
                    same = False
            f_info = f["/cellstate_model/run_info"]
            for key, val in info.items():
                if val != f_info[key][()]:
                    same = False
            if not (
                self.a.get_state_model().get_losses().cpu().numpy()
                == f["/cellstate_model/losses"]["losses"][()]
            ).all():
                same = False
        self.assertTrue(same)

    def test_hierarchy_assignment(self):
        self.a.fit_type(max_epochs=5, n_init=1, n_init_epochs=1)
        original_assignment = self.a.get_celltype_probabilities()
        hier_dict = self.a.get_hierarchy_dict()
        # expected_assignment = pd.DataFrame()
        # for key, cells in hier_dict.items():
        #     expected_assignment[key] = original_assignment[cells].sum(axis=1)
        actual_assignment = self.a.assign_celltype_hierarchy(depth=3)
        # self.assertTrue((original_assignment == actual_assignment).all().all())
        for cell in actual_assignment.columns:
            self.assertTrue(
                (actual_assignment[cell] == original_assignment[cell]).all()
            )

    def test_hdf5_load(self):
        hdf5_summary = "celltype_summary.hdf5"
        orig_ast = Astir(self.expr, self.marker_dict)
        orig_ast.fit_type(max_epochs=5, n_init=1, n_init_epochs=1)
        orig_ast.fit_state(max_epochs=5, n_init=1, n_init_epochs=1)
        orig_ast.save_models(hdf5_summary)
        new_ast = Astir()
        new_ast.load_model(hdf5_summary)

        orig_type_run_info = orig_ast.get_type_run_info()
        orig_state_run_info = orig_ast.get_state_run_info()
        new_type_run_info = new_ast.get_type_run_info()
        new_state_run_info = new_ast.get_state_run_info()
        for key, val in orig_type_run_info.items():
            if val != new_type_run_info[key]:
                raise AssertionError(
                    "variable "
                    + key
                    + " is different in original model and loaded model"
                )
        for key, val in orig_state_run_info.items():
            if val != new_state_run_info[key]:
                raise AssertionError(
                    "variable "
                    + key
                    + " is different in original model and loaded model"
                )

        orig_type_losses = orig_ast.get_type_losses()
        orig_state_losses = orig_ast.get_state_losses()
        new_type_losses = new_ast.get_type_losses()
        new_state_losses = new_ast.get_state_losses()
        if not (
            all(orig_type_losses == new_type_losses)
            and all(orig_state_losses == new_state_losses)
        ):
            raise AssertionError("loss is different in original model and loaded model")

    # def test_make_html(self):
    #     path = os.path.dirname(os.path.realpath(__file__)) + "/../../docs"
    #     cmd = ["make", "html"]
    #     p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, cwd=path)
    #     output, err = p.communicate(b"input data that is passed to subprocess' stdin")
    #     print(err)
    #     print(path)
    #     rc = p.returncode
    #     self.assertTrue(rc == 0)



from .data_readers import (
    from_anndata_yaml,
    from_csv_dir_yaml,
    from_csv_yaml,
    from_loompy_yaml,
)
from .scdataset import SCDataset

__all__ = [
    "from_csv_yaml",
    "from_csv_dir_yaml",
    "from_loompy_yaml",
    "from_anndata_yaml",
    "SCDataset",
]



import torch
import torch.nn as nn
import torch.nn.functional as F


class TypeRecognitionNet(nn.Module):
    """Type Recognition Neural Network.

    :param C: number of classes
    :param G: number of features
    :param hidden_size: size of hidden layers, defaults to 10
    """

    def __init__(self, C: int, G: int, hidden_size: int = 20) -> None:
        super(TypeRecognitionNet, self).__init__()
        self.hidden_1 = nn.Linear(G, hidden_size)
        self.hidden_2 = nn.Linear(hidden_size, C + 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """One forward pass.

        :param x: the input vector
        :return: the calculated cost value
        """
        x = self.hidden_1(x)
        x = F.leaky_relu(x)
        x = self.hidden_2(x)

        return F.softmax(x, dim=1), F.log_softmax(x, dim=1)



import os
import unittest
import warnings

import nbformat
import rootpath
from nbconvert.preprocessors import ExecutePreprocessor


def run_notebook(notebook_path):
    nb_name, _ = os.path.splitext(os.path.basename(notebook_path))

    with open(notebook_path) as f:
        nb = nbformat.read(f, as_version=4)

    # Configure the notebook execution mode
    proc = ExecutePreprocessor(timeout=600, kernel_name="python3")
    proc.allow_errors = True

    # Run the notebook
    root_path = rootpath.detect()
    path = root_path + "/docs/tutorials/notebooks"
    proc.preprocess(nb, {"metadata": {"path": path}})
    # Collect all errors
    errors = []
    for cell in nb.cells:
        if "outputs" in cell:
            for output in cell["outputs"]:
                if output.output_type == "error":
                    errors.append(output)

    return nb, errors


class TestNotebook(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestNotebook, self).__init__(*args, **kwargs)
        self.maxDiff = None

    def test_for_errors(self):
        warnings.filterwarnings("ignore")
        root_path = rootpath.detect()
        dirname = os.path.join(root_path, "docs/tutorials/notebooks")

        nb_names = [
            os.path.join(dirname, fn)
            for fn in os.listdir(dirname)
            if os.path.splitext(fn)[1] == ".ipynb"
        ]

        for fn in nb_names:
            _, errors = run_notebook(fn)
            self.assertEqual(errors, [], "Unexpected error in {}".format(fn))


if __name__ == "__main__":
    unittest.main()



import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader, Dataset


class SCDataset(Dataset):
    """Container for single-cell proteomic data in the form of
    a pytorch dataset

    :param expr_input: Input expression data. See details :`expr_input` is either a `pd.DataFrame`
        or a three-element `tuple`. When it is `pd.DataFrame`, its index and column should indicate the cell
        name and feature name of the dataset; when it is a three-element `tuple`, it should be in the form
        of `Tuple[Union[np.array, torch.Tensor], List[str], List[str]]` and  its first element should
        be the actual dataset as either `np.array` or `torch.tensor`, the second element should be
        a list containing the name of the columns or the names of features, the third element should be a
        list containing the name of the indices or the names of the cells.:
    :param marker_dict: Marker dictionary containing cell type and
        information. See details :The dictionary maps the name of cell type/state to protein features. :
    :param design: A design matrix
    :param include_other_column: Should an additional 'other' column be included?
    :param dtype: torch datatype of the model
    """

    def __init__(
        self,
        expr_input: Union[
            pd.DataFrame, Tuple[Union[np.array, torch.Tensor], List[str], List[str]]
        ],
        marker_dict: Dict[str, List[str]],
        include_other_column: bool,
        design: Optional[Union[np.array, pd.DataFrame]] = None,
        dtype: torch.dtype = torch.float64,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        """Initialize an SCDataset object."""
        self._dtype = dtype
        self._marker_dict = marker_dict
        self._m_features = sorted(
            list(set([l for s in marker_dict.values() for l in s]))
        )
        self._classes = list(marker_dict.keys())

        self._device = device
        ## sanitize features
        if len(self._classes) <= 1 and include_other_column:
            raise NotClassifiableError(
                "Classification failed. There "
                + "should be at least two cell classes to classify the data into."
            )
        self._marker_mat = self._construct_marker_mat(
            include_other_column=include_other_column
        )

        if isinstance(expr_input, pd.DataFrame):
            self._exprs = self._process_df_input(expr_input)
            self._expr_features = list(expr_input.columns)
            self._cell_names = list(expr_input.index)
        elif isinstance(expr_input, tuple):
            self._expr_features = expr_input[1]
            self._cell_names = expr_input[2]
            self._exprs = self._process_tp_input(expr_input[0])
        self._design = self._fix_design(design)
        ## sanitize df
        if self._exprs.shape[0] <= 0:
            raise NotClassifiableError(
                "Classification failed. There "
                + "should be at least one row of data to be classified."
            )
        self._exprs_mean = self._exprs.mean(0)
        self._exprs_std = self._exprs.std(0)

    def _process_df_input(self, df_input: pd.DataFrame) -> torch.Tensor:
        """Processes input as pd.DataFrame and convert it into torch.Tensor

        :param df_input: the input
        :raises NotClassifiableError: raised when there is no overlap between the
            data and the marker
        :return: the processed input as a torch.Tensor
        """
        try:
            Y_np = df_input[self._m_features].values
            return torch.from_numpy(Y_np).to(device=self._device, dtype=self._dtype)
        except (KeyError):
            raise NotClassifiableError(
                "Classification failed. There's no "
                + "overlap between marked features and expression features for "
                + "the classification of cell type/state."
            )

    def _process_tp_input(self, in_data: Union[torch.Tensor, np.array]) -> torch.Tensor:
        """Process the input as Tuple[np.array, np.array, np.array] and convert it
            to torch.Tensor.

        :param in_data: input as a np.array or torch.tensor
        :raises NotClassifiableError: raised when there is no overlap between marked
            features and expression feature.
        :return: the processed input as a torch.Tensor
        """
        ind = [
            self._expr_features.index(name)
            for name in self._m_features
            if name in self._expr_features
        ]
        if len(ind) <= 0:
            raise NotClassifiableError(
                "Classification failed. There's no "
                + "overlap between marked features and expression features for "
                + "the classification of cell type/state."
            )
        if len(ind) < len(self._m_features):
            warnings.warn("Classified features are less than marked features.")
        Y_np = in_data[:, ind]
        if torch.is_tensor(Y_np):
            return Y_np.to(device=self._device, dtype=self._dtype)
        return torch.from_numpy(Y_np).to(device=self._device, dtype=self._dtype)

    def _construct_marker_mat(self, include_other_column: bool) -> torch.Tensor:
        """Construct a marker matrix.

        :param include_other_column: indicates whether or not include other columns.
        :return: A marker matrix. The rows are features and the coloumns are
            the corresponding classes (type/state).
        """
        G = self.get_n_features()
        C = self.get_n_classes()

        marker_mat = torch.zeros(
            (G, C + 1 if include_other_column else C), dtype=self._dtype
        ).to(self._device)
        for g, feature in enumerate(self._m_features):
            for c, cell_class in enumerate(self._classes):
                if feature in self._marker_dict[cell_class]:
                    marker_mat[g, c] = 1.0
        return marker_mat

    def __len__(self) -> int:
        """Length of the input file

        :return: total number of cells
        """
        # N
        return self._exprs.shape[0]

    def __getitem__(
        self, idx: Union[slice, int]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Returns the protein expression of the indexed cell on the SCDataset
        object

        :param idx: the index of the cell
        :return: raw protein expression, normalized protein expression,
        sanitized design matrix of the cell at index
        """
        y = self._exprs[idx, :]
        x = (y - self._exprs_mean) / self._exprs_std
        return y, x, self._design[idx, :]

    def _fix_design(self, design: Union[np.array, pd.DataFrame]) -> torch.Tensor:
        """Sanitize the design matrix.

        :param design: the unsanitized design matrix
        :raises NotClassifiableError: raised when the design matrix has
            different number of rows from the expression data
        :return: the sanitized design matrix
        """
        d = None
        if design is None:
            d = torch.ones((self._exprs.shape[0], 1)).to(
                device=self._device, dtype=self._dtype
            )
        else:
            if isinstance(design, pd.DataFrame):
                design = design.values
            d = torch.from_numpy(design).to(device=self._device, dtype=self._dtype)

        if d.shape[0] != self._exprs.shape[0]:
            raise NotClassifiableError(
                "Number of rows of design matrix "
                + "must equal number of rows of expression data"
            )
        return d

    def rescale(self) -> None:
        """Normalize the expression data."""
        self._exprs = self._exprs / (self.get_sigma())

    def get_dtype(self) -> torch.dtype:
        """Get the dtype of the `SCDataset`.

        :return: `self._dtype`
        :rtype: torch.dtype
        """
        return self._dtype

    def get_exprs(self) -> torch.Tensor:
        """Return the expression data as a :class:`torch.Tensor`."""
        return self._exprs

    def get_exprs_df(self) -> pd.DataFrame:
        """Return the expression data as a :class:`pandas.DataFrame`."""
        df = pd.DataFrame(self._exprs.detach().cpu().numpy())
        df.index = self.get_cell_names()
        df.columns = self.get_features()
        return df

    def get_marker_mat(self) -> torch.Tensor:
        """Return the marker matrix as a :class:`torch.Tensor`."""
        return self._marker_mat

    def get_mu(self) -> torch.Tensor:
        """Get the mean expression of each protein as a :class:`torch.Tensor`."""
        return self._exprs_mean

    def get_sigma(self) -> torch.Tensor:
        """Get the standard deviation of each protein

        :return: standard deviation of each protein
        """
        return self._exprs_std

    def get_n_classes(self) -> int:
        """Get the number of 'classes': either the number of cell types or cell states."""
        return len(self._classes)

    def get_n_cells(self) -> int:
        """Get the number of cells: either the number of cell types or cell states."""
        return len(self.get_cell_names())

    def get_n_features(self) -> int:
        """Get the number of features (proteins)."""
        return len(self._m_features)

    def get_features(self) -> List[str]:
        """Get the features (proteins).

        :return: return self._m_features
        :rtype: List[str]
        """
        return self._m_features

    def get_cell_names(self) -> List[str]:
        """Get the cell names.

        :return: return self._cell_names
        :rtype: List[str]
        """
        return self._cell_names

    def get_classes(self) -> List[str]:
        """Get the cell types/states.

        :return: return self._classes
        :rtype: List[str]
        """
        return self._classes

    def get_design(self) -> torch.Tensor:
        """Get the design matrix.

        :return: return self._design
        :rtype: torch.Tensor
        """
        return self._design

    def normalize(
        self,
        percentile_lower: float = 0,
        percentile_upper: float = 99.9,
        cofactor: float = 5.0,
    ) -> None:
        """Normalize the expression data

        This performs a two-step normalization:
        1. A `log(1+x)` transformation to the data
        2. Winsorizes to (`percentile_lower`, `percentile_upper`)

        :param percentile_lower: the lower bound percentile for
            winsorization, defaults to 0
        :param percentil_upper: the upper bound percentile for winsorization,
            defaults to 99.9
        :param cofactor: a cofactor constant, defaults to 5.0
        """

        with torch.no_grad():
            exprs = self.get_exprs().cpu().numpy()
            exprs = np.arcsinh(exprs / cofactor)
            q_low = np.percentile(exprs, (percentile_lower), axis=0)
            q_high = np.percentile(exprs, (percentile_upper), axis=0)

            for g in range(exprs.shape[1]):
                exprs[:, g][exprs[:, g] < q_low[g]] = q_low[g]
                exprs[:, g][exprs[:, g] > q_high[g]] = q_high[g]

            self._exprs = torch.tensor(exprs)

    def get_mu_init(self, n_putative_cells: int = 10) -> np.ndarray:
        """Intelligent initialization for mu parameters

        See manuscript for details

        :param n_putative_cells: Number of cells to guess as given cell type
        """
        df_exprs = self.get_exprs_df()

        df_scaled = df_exprs.copy()
        scaler = StandardScaler()
        df_scaled[df_scaled.columns] = scaler.fit_transform(
            df_scaled[df_scaled.columns]
        )
        df_scaled.head()

        putative_cells = {}

        for cell_type, type_markers in self._marker_dict.items():
            scored_by_marker = df_scaled[type_markers].mean(1)
            putative_cells[cell_type] = scored_by_marker.nlargest(
                n_putative_cells
            ).index

        mean_inits = {}

        for feature in self._m_features:
            ## List of cell types that do not contain feature
            celltypes_no_feature = [
                c for c in self._classes if not feature in self._marker_dict[c]
            ]
            indices_to_use = [putative_cells[ct] for ct in celltypes_no_feature]
            idx = indices_to_use[0]
            for i in range(1, len(indices_to_use)):
                idx = idx.union(indices_to_use[i])
            mean_inits[feature] = df_exprs.loc[idx][feature].mean()

        mean_init = pd.Series(mean_inits).to_numpy()

        return mean_init


class NotClassifiableError(RuntimeError):
    pass



from .abstract import AstirModel
from .cellstate import CellStateModel
from .cellstate_recognet import StateRecognitionNet
from .celltype import CellTypeModel
from .celltype_recognet import TypeRecognitionNet

__all__ = [
    "CellTypeModel",
    "CellStateModel",
    "AstirModel",
    "TypeRecognitionNet",
    "StateRecognitionNet",
]



"""
Cell State Model
"""
import warnings
from collections import OrderedDict
from typing import Dict, Generator, List, Optional, Tuple, Union

import h5py
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from tqdm import trange

from astir.data import SCDataset

from .abstract import AstirModel
from .cellstate_recognet import StateRecognitionNet


class CellStateModel(AstirModel):
    """Class to perform statistical inference to on the activation
        of states (pathways) across cells

    :param dset: the input gene expression dataset, defaults to None
    :param const: See parameter ``const`` in
        :func:`astir.models.StateRecognitionNet`, defaults to 2
    :param dropout_rate: See parameter ``dropout_rate`` in
        :func:`astir.models.StateRecognitionNet`, defaults to 0
    :param batch_norm: See parameter ``batch_norm`` in
        :func:`astir.models.StateRecognitionNet`, defaults to False
    :param random_seed: the random seed number to reproduce results, defaults to 42
    :param dtype: torch datatype to use in the model, defaults to torch.float64
    :param device: torch.device's cpu or gpu, defaults to torch.device("cpu")
    """

    def __init__(
        self,
        dset: SCDataset = None,
        const: int = 2,
        dropout_rate: float = 0,
        batch_norm: bool = False,
        random_seed: int = 42,
        dtype: torch.dtype = torch.float64,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        super().__init__(dset, random_seed, dtype, device)

        # Setting random seeds
        self.random_seed = random_seed
        torch.manual_seed(self.random_seed)
        np.random.seed(self.random_seed)

        self._optimizer: Optional[torch.optim.Adam] = None
        self.const, self.dropout_rate, self.batch_norm = const, dropout_rate, batch_norm
        if self._dset is not None:
            self._param_init()

        # Convergence flag
        self._is_converged = False

    def _param_init(self) -> None:
        """Initializes sets of parameters"""
        if self._dset is None:
            raise Exception("the dataset is not provided")
        N = len(self._dset)
        C = self._dset.get_n_classes()
        G = self._dset.get_n_features()

        initializations = {
            "log_sigma": torch.log(self._dset.get_sigma().mean()),
            "mu": torch.reshape(self._dset.get_mu(), (1, -1)),
        }

        # Include beta or not
        d = torch.distributions.Uniform(
            torch.tensor(0.0, dtype=self._dtype), torch.tensor(1.5, dtype=self._dtype)
        )
        initializations["log_w"] = torch.log(d.sample((C, G)))

        self._variables = {
            n: i.to(self._device).detach().clone().requires_grad_()
            for (n, i) in initializations.items()
        }

        self._data = {
            "rho": self._dset.get_marker_mat().T.to(self._device),
        }

        self._recog = StateRecognitionNet(
            C,
            G,
            const=self.const,
            dropout_rate=self.dropout_rate,
            batch_norm=self.batch_norm,
        ).to(device=self._device, dtype=self._dtype)

    def load_hdf5(self, hdf5_name: str) -> None:
        """Initializes Cell State Model from a hdf5 file type

        :param hdf5_name: file path
        """
        self._assignment = pd.read_hdf(
            hdf5_name, "cellstate_model/cellstate_assignments"
        )
        with h5py.File(hdf5_name, "r") as f:
            grp = f["cellstate_model"]
            param = grp["parameters"]
            self._variables = {
                "mu": torch.tensor(np.array(param["mu"])),
                "log_sigma": torch.tensor(np.array(param["log_sigma"])),
                "log_w": torch.tensor(np.array(param["log_w"])),
            }
            self._data = {"rho": torch.tensor(np.array(param["rho"]))}
            self._losses = torch.tensor(np.array(grp["losses"]["losses"]))

            rec = grp["recog_net"]
            hidden1_W = torch.tensor(np.array(rec["linear1.weight"]))
            hidden2_W = torch.tensor(np.array(rec["linear2.weight"]))
            hidden3_mu_W = torch.tensor(np.array(rec["linear3_mu.weight"]))
            hidden3_std_W = torch.tensor(np.array(rec["linear3_std.weight"]))
            state_dict = {
                "linear1.weight": hidden1_W,
                "linear1.bias": torch.tensor(np.array(rec["linear1.bias"])),
                "linear2.weight": hidden2_W,
                "linear2.bias": torch.tensor(np.array(rec["linear2.bias"])),
                "linear3_mu.weight": hidden3_mu_W,
                "linear3_mu.bias": torch.tensor(np.array(rec["linear3_mu.bias"])),
                "linear3_std.weight": hidden3_std_W,
                "linear3_std.bias": torch.tensor(np.array(rec["linear3_std.bias"])),
            }
            state_dict = OrderedDict(state_dict)
            self._recog = StateRecognitionNet(
                hidden3_mu_W.shape[0],
                hidden1_W.shape[1],
                const=self.const,
                dropout_rate=self.dropout_rate,
                batch_norm=self.batch_norm,
            ).to(device=self._device, dtype=self._dtype)
            self._recog.load_state_dict(state_dict)
            self._recog.eval()

    def _loss_fn(
        self,
        mu_z: torch.Tensor,
        std_z: torch.Tensor,
        z_sample: torch.Tensor,
        y_in: torch.Tensor,
    ) -> torch.Tensor:
        """Returns the calculated loss

        :param mu_z: the predicted mean of z
        :param std_z: the predicted standard deviation of z
        :param z_sample: the sampled z values
        :param y_in: the input data
        :return: the loss
        """
        S = y_in.shape[0]

        # log posterior q(z) approx p(z|y)
        q_z_dist = torch.distributions.Normal(loc=mu_z, scale=torch.exp(std_z))
        log_q_z = q_z_dist.log_prob(z_sample)

        # log likelihood p(y|z)
        rho_w = torch.mul(self._data["rho"], torch.exp(self._variables["log_w"]))
        mean = self._variables["mu"] + torch.matmul(z_sample, rho_w)
        std = torch.exp(self._variables["log_sigma"]).reshape(1, -1)
        p_y_given_z_dist = torch.distributions.Normal(loc=mean, scale=std)
        log_p_y_given_z = p_y_given_z_dist.log_prob(y_in)

        # log prior p(z)
        p_z_dist = torch.distributions.Normal(0, 1)
        log_p_z = p_z_dist.log_prob(z_sample)

        loss = (1 / S) * (
            torch.sum(log_q_z) - torch.sum(log_p_y_given_z) - torch.sum(log_p_z)
        )

        return loss

    def _forward(
        self,
        Y: Optional[torch.Tensor],
        X: Optional[torch.Tensor] = None,
        design: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """One forward pass

        :param Y: dataset to do forward pass on
        :return: mu_z, std_z, z_sample
        """
        mu_z, std_z = self._recog(Y)

        std = torch.exp(std_z)
        eps = torch.randn_like(std)
        z_sample = eps * std + mu_z

        return mu_z, std_z, z_sample

    def fit(
        self,
        max_epochs: int = 50,
        learning_rate: float = 1e-3,
        batch_size: int = 128,
        delta_loss: float = 1e-3,
        delta_loss_batch: int = 10,
        msg: str = "",
    ) -> None:
        """ Runs train loops until the convergence reaches delta_loss for\
            delta_loss_batch sizes or for max_epochs number of times

        :param max_epochs: number of train loop iterations, defaults to 50
        :param learning_rate: the learning rate, defaults to 0.01
        :param batch_size: the batch size, defaults to 128
        :param delta_loss: stops iteration once the loss rate reaches\
            delta_loss, defaults to 0.001
        :param delta_loss_batch: the batch size to consider delta loss,\
            defaults to 10
        :param msg: iterator bar message, defaults to empty string
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")

        # Returns early if the model has already converged
        if self._is_converged:
            return

        # Create an optimizer if there is no optimizer
        if self._optimizer is None:
            opt_params = list(self._recog.parameters())
            opt_params += list(self._variables.values())  # type: ignore
            self._optimizer = torch.optim.Adam(opt_params, lr=learning_rate)

        iterator = trange(
            max_epochs,
            desc="training restart" + msg,
            unit="epochs",
            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{rate_fmt}{postfix}]",
        )
        train_iterator = DataLoader(
            self._dset, batch_size=min(batch_size, len(self._dset))
        )
        for ep in iterator:
            for i, (y_in, x_in, _) in enumerate(train_iterator):
                self._optimizer.zero_grad()

                mu_z, std_z, z_samples = self._forward(
                    x_in.type(self._dtype).to(self._device)
                )

                loss = self._loss_fn(
                    mu_z, std_z, z_samples, x_in.type(self._dtype).to(self._device)
                )

                loss.backward()

                self._optimizer.step()

            loss_detached = loss.cpu().detach().item()

            self._losses = torch.cat(
                (self._losses, torch.tensor([loss_detached], dtype=self._dtype))
            )

            if len(self._losses) > delta_loss_batch:
                curr_mean = torch.mean(self._losses[-delta_loss_batch:])
                prev_mean = torch.mean(self._losses[-delta_loss_batch - 1 : -1])
                curr_delta_loss = (prev_mean - curr_mean) / prev_mean
                delta_cond_met = 0 <= curr_delta_loss.item() < delta_loss
            else:
                delta_cond_met = False

            iterator.set_postfix_str("current loss: " + str(round(loss_detached, 1)))

            if delta_cond_met:
                self._is_converged = True
                iterator.close()
                break

        g = self.get_final_mu_z().detach().cpu().numpy()
        self._assignment = pd.DataFrame(g)
        self._assignment.columns = self._dset.get_classes()
        self._assignment.index = self._dset.get_cell_names()

    def get_recognet(self) -> StateRecognitionNet:
        """Getter for the recognition net

        :return: the recognition net
        """
        return self._recog

    def get_final_mu_z(self, new_dset: Optional[SCDataset] = None) -> torch.Tensor:
        """Returns the mean of the predicted z values for each core

        :param new_dset: returns the predicted z values of this dataset on
            the existing model. If None, it predicts using the existing
            dataset, defaults to None
        :return: the mean of the predicted z values for each core
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        if new_dset is None:
            _, x_in, _ = self._dset[:]  # should be the scaled
            # one
        else:
            _, x_in, _ = new_dset[:]
        final_mu_z, _, _ = self._forward(x_in.type(self._dtype).to(self._device))

        return final_mu_z

    def get_correlations(self) -> np.array:
        """Returns a C (# of pathways) X G (# of proteins) matrix
        where each element represents the correlation value of the pathway
        and the protein

        :return: matrix of correlation between all pathway and protein pairs.
        """
        if self._dset is None:
            raise Exception("No dataset input to the model")

        state_assignment = self.get_final_mu_z().detach().cpu().numpy()
        y_in = self._dset.get_exprs()

        feature_names = self._dset.get_features()
        state_names = self._dset.get_classes()
        G = self._dset.get_n_features()
        C = self._dset.get_n_classes()
        corr_mat = np.zeros((C, G))
        # Make a matrix of correlations between all states and proteins
        for c, state in enumerate(state_names):
            for g, feature in enumerate(feature_names):
                states = state_assignment[:, c]
                protein = y_in[:, g].cpu()
                corr_mat[c, g] = np.corrcoef(protein, states)[0, 1]

        return corr_mat

    def diagnostics(self) -> pd.DataFrame:
        """Run diagnostics on cell type assignments

        See :meth:`astir.Astir.diagnostics_cellstate` for full documentation
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        feature_names = self._dset.get_features()
        state_names = self._dset.get_classes()

        corr_mat = self.get_correlations()

        # Correlation values of all marker proteins
        marker_mat = self._dset.get_marker_mat().T.cpu().numpy()
        marker_corr = marker_mat * corr_mat
        marker_corr[marker_mat == 0] = np.inf

        # Smallest correlation values for each pathway
        min_marker_corr = np.min(marker_corr, axis=1).reshape(-1, 1)
        min_marker_proteins = np.take(feature_names, np.argmin(marker_corr, axis=1))

        # Correlation values of all non marker proteins
        non_marker_mat = 1 - self._dset.get_marker_mat().T.cpu().numpy()
        non_marker_corr = non_marker_mat * corr_mat
        non_marker_corr[non_marker_mat == 0] = -np.inf

        # Any correlation values where non marker proteins is greater than
        # the smallest correlation values of marker proteins
        bad_corr_marker = np.array(non_marker_corr > min_marker_corr, dtype=np.int32)

        # Problem summary
        indices = np.argwhere(bad_corr_marker > 0)

        col_names = [
            "pathway",
            "protein A",
            "correlation of protein A",
            "protein B",
            "correlation of protein B",
            "note",
        ]

        problems = []
        for index in indices:
            state_index = index[0]
            protein_index = index[1]
            state = state_names[index[0]]
            marker_protein = min_marker_proteins[state_index]
            non_marker_protein = feature_names[protein_index]
            problem = {
                "pathway": state,
                "marker_protein": marker_protein,
                "corr_of_marker_protein": min_marker_corr[state_index][0],
                "non_marker_protein": non_marker_protein,
                "corr_of_non_marker_protein": non_marker_corr[
                    state_index, protein_index
                ],
                "msg": "{} is marker for {} but {} isn't".format(
                    marker_protein, state, non_marker_protein
                ),
            }
            problems.append(problem)

        if len(problems) > 0:
            df_issues = pd.DataFrame(problems)
            df_issues.columns = col_names
        else:
            df_issues = pd.DataFrame(columns=col_names)

        return df_issues


class NotClassifiableError(RuntimeError):
    """Raised when the input data is not classifiable."""

    pass



""" Tests the following for SCDataset:
- Whether all methods have docstrings
- Whether all methods contain type hints and return type hints
- Whether return type hints are correct
- Any other code smells like a subclass has the same method definitions as
its superclass
"""
import unittest

from astir import Astir
from astir.data import SCDataset
from astir.models import (
    AstirModel,
    CellStateModel,
    CellTypeModel,
    StateRecognitionNet,
    TypeRecognitionNet,
)


class TestCodeSmells(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestCodeSmells, self).__init__(*args, **kwargs)
        self.classes = [
            Astir,
            CellStateModel,
            CellTypeModel,
            StateRecognitionNet,
            TypeRecognitionNet,
            AstirModel,
            SCDataset,
        ]
        self.class_paths = [
            "astir/astir.py",
            "astir/models/cellstate.py",
            "astir/models/celltype.py",
            "astir/models/cellstate_recognet.py",
            "astir/models/celltype_recognet.py",
            "astir/models/abstract.py",
            "astir/data/scdataset.py",
        ]

    def test_docstrings_exists_all_methods(self):
        for cl in self.classes:
            cl_name = cl.__name__
            class_dict = dict(cl.__dict__)

            class_dict.pop("__module__")
            class_dict.pop("__init__")

            no_doc = []
            for key, value in class_dict.items():
                if value.__doc__ is None:
                    no_doc.append(key)

            err_msg = "{} has methods without docstring: ".format(cl_name)
            for method in no_doc:
                err_msg += method + ", "
            self.assertTrue(no_doc == [], err_msg[:-2])

    def test_type_hints_exist_all_methods(self):
        from typing import get_type_hints
        from inspect import signature

        for cl in self.classes:
            cl_name = cl.__name__
            class_dict = dict(cl.__dict__)

            class_dict.pop("__module__")
            class_dict.pop("__doc__")

            no_type_hint = []
            no_return_hint = []
            for key, value in class_dict.items():
                try:
                    type_hints = get_type_hints(value)
                except:
                    continue

                if not type_hints.__contains__("return"):
                    no_return_hint.append(key)
                else:
                    type_hints.pop("return")
                params = list(signature(value).parameters)
                params.remove("self")

                if len(params) != len(type_hints):
                    no_type_hint.append(key)

            err_msg_type_hint = "{} is missing type hints for methods: " "".format(
                cl_name
            )
            for method in no_type_hint:
                err_msg_type_hint += method + ", "

            err_msg_return_hint = "{} is missing return hints for " "methods: ".format(
                cl_name
            )
            for method in no_return_hint:
                err_msg_return_hint += method + ", "
            self.assertTrue(no_type_hint == [], err_msg_type_hint[:-2])
            self.assertTrue(no_return_hint == [], err_msg_return_hint[:-2])

    def test_correct_return_types_and_detect_code_smells(self):
        import subprocess, rootpath, os

        for cl_path in self.class_paths:
            root_dir = rootpath.detect()
            path = os.path.join(root_dir, cl_path)

            process = subprocess.Popen(
                ["mypy", path, "--ignore-missing-imports", "--no-site-packages"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            stdout, stderr = process.communicate()
            stdout = stdout.decode("utf-8")
            stderr = stderr.decode("utf-8")
            output = str(stdout).split("\n")

            errors = []
            for line in output:
                if line.__contains__("error:"):
                    errors.append(line)

            self.assertTrue(
                errors == [],
                "Following errors were produced by "
                "MyPy: \n{}".format("\n".join(errors)),
            )
            self.assertTrue(
                stderr == "", "Following error were produced by MyPy: {}".format(stderr)
            )

    def test_data_reader_code_smell(self):
        from typing import get_type_hints
        from inspect import signature
        from astir.data.data_readers import (
            from_csv_yaml,
            from_csv_dir_yaml,
            from_loompy_yaml,
            from_anndata_yaml,
        )

        funcs = [from_csv_yaml, from_csv_dir_yaml, from_loompy_yaml, from_anndata_yaml]

        # Test whether all of them have type hints
        param_msgs = []
        docstring_msgs = []
        for func in funcs:
            type_hints = get_type_hints(func)
            param_with_hints = list(type_hints.keys())
            all_params = list(signature(func).parameters)

            param_diff = list(set(all_params) - set(param_with_hints))

            for param in param_diff:
                param_msgs.append(
                    "astir.data.data_readers.{} needs type hint "
                    "for parameter {}".format(func.__name__, param)
                )

            # Docstring test
            if func.__doc__ is None:
                docstring_msgs.append(
                    "astir.data.data_readers.{} needs a "
                    "docstring".format(func.__name__)
                )

        self.assertTrue(param_msgs == [], "\n".join(param_msgs))
        self.assertTrue(docstring_msgs == [], "\n".join(docstring_msgs))


if __name__ == "__main__":
    unittest.main()



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="astir",
    version="0.1.3",
    author="Jinyu Hou, Sunyun Lee, Michael Geuenich, Kieran Campbell",
    author_email="kierancampbell@lunenfeld.ca",
    description=" ",
    long_description=long_description,
    # long_description_content_type="text/markdown",
    url="https://github.com/camlab-bioml/astir",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="GPLv2",
    install_requires=[
        "torch",
        "pandas",
        "numpy",
        "nbformat",
        "pyyaml",
        "sklearn",
        "argparse",
        "matplotlib",
        "loompy",
        "tqdm",
        "anndata",
        "rootpath",
        "nbconvert",
        "nbformat",
        "seaborn",
        "jupyter_client",
        "ipykernel",
        "ipython",
        "h5py",
        "FlowCytometryTools",
        "fastcluster",
        "sphinx",
        "sphinx_autodoc_typehints",
        "sphinx_rtd_theme",
        "autodocsumm",
        "nbsphinx",
        "nbsphinx_link",
        "tables",
        "typeguard",
        "mypy",
    ],
    project_urls={
        "Documentation": "https://astir.readthedocs.io/en/latest/",
        "Source Code": "https://github.com/camlab-bioml/astir",
    },
    include_package_data=True,
    zip_safe=False,
    # test_suite="nose.collector",
    # test_require=["nose"],
    scripts=["bin/astir"],
)



# astir: Automated aSsignmenT sIngle-cell pRoteomics

# VSCode tips for python:
# ctrl+` to open terminal
# cmd+P to go to file
import os
import re
import warnings
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from sklearn.preprocessing import StandardScaler

from .data.scdataset import SCDataset
from .models.cellstate import CellStateModel
from .models.celltype import CellTypeModel


class Astir:
    """Create an Astir object

    :param input_expr: the single cell protein expression dataset
    :type input_expr: Union[pd.DataFrame, Tuple[np.array, List[str], List[str]], Tuple[SCDataset, SCDataset]]
    :param marker_dict: the marker dictionary which maps cell type/state to protein features, defaults to None
    :type marker_dict: Dict[str, Dict[str, str]], optional
    :param design: the design matrix labeling the grouping of cell, defaults to None
    :type design: Union[pd.DataFrame, np.array], optional
    :param random_seed: random seed for parameter initialization, defaults to `1234`
    :type random_seed: int, optional
    :param dtype: dtype of data, defaults to `torch.float64`
    :type dtype: torch.dtype, optional
    :raises NotClassifiableError: raised if the model is not trainable
    """

    def __init__(
        self,
        input_expr: Union[
            pd.DataFrame,
            Tuple[np.array, List[str], List[str]],
            Tuple[SCDataset, SCDataset],
        ] = None,
        marker_dict: Optional[Dict[str, Dict[str, List[str]]]] = None,
        design: Union[pd.DataFrame, np.array, None] = None,
        random_seed: int = 1234,
        dtype: torch.dtype = torch.float64,
    ) -> None:

        if not isinstance(random_seed, int):
            raise NotClassifiableError("Random seed is expected to be an integer.")
        torch.manual_seed(random_seed)
        self.random_seed = random_seed

        if dtype != torch.float32 and dtype != torch.float64:
            raise NotClassifiableError(
                "Dtype must be one of torch.float32 and torch.float64."
            )
        self._dtype = dtype

        self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self._type_ast: Optional[CellTypeModel] = None
        self._state_ast: Optional[CellStateModel] = None
        self._type_run_info: dict = {}
        self._state_run_info: dict = {}

        self._type_dset: Optional[SCDataset] = None
        self._state_dset: Optional[SCDataset] = None

        self._hierarchy_dict: Optional[Dict[str, List[str]]] = None

        type_dict, state_dict, self._hierarchy_dict = self._sanitize_dict(marker_dict)
        if isinstance(input_expr, tuple) and len(input_expr) == 2:
            self._type_dset, self._state_dset = input_expr[0], input_expr[1]
        else:
            if type_dict is not None:
                self._type_dset = SCDataset(
                    expr_input=input_expr,
                    marker_dict=type_dict,
                    design=design,
                    include_other_column=True,
                    dtype=self._dtype,
                    device=self._device,
                )
            if state_dict is not None:
                self._state_dset = SCDataset(
                    expr_input=input_expr,
                    marker_dict=state_dict,
                    design=design,
                    include_other_column=False,
                    dtype=self._dtype,
                    device=self._device,
                )

    def _sanitize_dict(
        self, marker_dict: Optional[Dict[str, Dict[str, List[str]]]]
    ) -> Union[List[None], List[dict]]:
        """Sanitizes the marker dictionary.

        :param marker_dict: dictionary read from the yaml file
        :raises NotClassifiableError: raized when the marker dictionary doesn't
             have required format
        :return: dictionaries, the first is the cell type dict, the second is the cell state
            dict and the third is the hierarchy dict.
        """
        dics = [None, None, None]
        if marker_dict is not None:
            ct = re.compile("cell[^a-zA-Z0-9]*type", re.IGNORECASE)
            cs = re.compile("cell[^a-zA-Z0-9]*state", re.IGNORECASE)
            h = re.compile(".*hierarch.*", re.IGNORECASE)

            def interpret(key: str, dic: Any) -> None:
                if ct.match(key):
                    dics[0] = dic
                elif cs.match(key):
                    dics[1] = dic
                elif h.match(key):
                    dics[2] = dic

            for key, dic in marker_dict.items():
                interpret(key, dic)

        return dics

    def fit_type(
        self,
        max_epochs: int = 50,
        learning_rate: float = 1e-3,
        batch_size: int = 128,
        delta_loss: float = 1e-3,
        n_init: int = 5,
        n_init_epochs: int = 5,
    ) -> None:
        """Run Variational Bayes to infer cell types

        :param max_epochs: maximum number of epochs to train
        :param learning_rate: ADAM optimizer learning rate
        :param batch_size: minibatch size
        :param delta_loss: stops iteration once the loss rate reaches
            delta_loss, defaults to `0.001`
        :param n_inits: number of random initializations
        """
        if self._type_dset is None:
            raise NotClassifiableError(
                "Dataset or marker for cell type classification is not provided"
            )
        np.random.seed(self.random_seed)
        seeds = np.random.randint(1, 100000000, n_init)
        type_models = [
            CellTypeModel(
                self._type_dset,
                int(seed),
                self._dtype,
            )
            for seed in seeds
        ]
        n_init_epochs = min(max_epochs, n_init_epochs)
        for i in range(n_init):
            type_models[i].fit(
                n_init_epochs,
                learning_rate,
                batch_size,
                delta_loss,
                msg=" " + str(i + 1) + "/" + str(n_init),
            )

        losses = torch.tensor([m.get_losses()[-1] for m in type_models])

        best_ind = int(torch.argmin(losses).item())
        self._type_ast = type_models[best_ind]
        self._type_ast.fit(
            max_epochs, learning_rate, batch_size, delta_loss, msg=" (final)"
        )

        if not self._type_ast.is_converged():
            msg = (
                "Maximum epochs reached. More iteration may be needed to"
                + " complete the training."
            )
            warnings.warn(msg)

        self._type_run_info = {
            "max_epochs": max_epochs,
            "learning_rate": learning_rate,
            "batch_size": batch_size,
            "delta_loss": delta_loss,
            "n_init": n_init,
            "n_init_epochs": n_init_epochs,
        }
        return None

    def fit_state(
        self,
        max_epochs: int = 50,
        learning_rate: float = 1e-3,
        batch_size: int = 128,
        delta_loss: float = 1e-3,
        n_init: int = 5,
        n_init_epochs: int = 5,
        delta_loss_batch: int = 10,
        const: int = 2,
        dropout_rate: float = 0,
        batch_norm: bool = False,
    ) -> None:
        """Run Variational Bayes to infer cell states

        :param max_epochs: number of epochs, defaults to `100`
        :param learning_rate: the learning rate, defaults to `1e-2`
        :param n_init: the number of initial parameters to compare,
            defaults to `5`
        :param delta_loss: stops iteration once the loss rate reaches
            delta_loss, defaults to `0.001`
        :param delta_loss_batch: the batch size to consider delta loss,
            defaults to `10`
        """
        if self._state_dset is None:
            raise NotClassifiableError(
                "Dataset or marker for cell state classification is not provided"
            )
        cellstate_models = []
        cellstate_losses = []

        if delta_loss_batch >= max_epochs:
            warnings.warn("Delta loss batch size is greater than the number of epochs")

        if n_init_epochs > max_epochs:
            warnings.warn(
                "n_init_epochs cannot be greater than the "
                "max_epochs. Models will be trained with max_epochs."
            )

        for i in range(n_init):
            # Initializing a model
            model = CellStateModel(
                const=const,
                dropout_rate=dropout_rate,
                batch_norm=batch_norm,
                dset=self._state_dset,
                random_seed=(self.random_seed + i),
                dtype=self._dtype,
            )
            # Fitting the model
            n_init_epochs = min(max_epochs, n_init_epochs)
            model.fit(
                max_epochs=n_init_epochs,
                learning_rate=learning_rate,
                batch_size=batch_size,
                delta_loss=delta_loss,
                delta_loss_batch=delta_loss_batch,
                msg=" " + str(i + 1) + "/" + str(n_init),
            )
            loss = model.get_losses()[0:n_init_epochs]

            cellstate_losses.append(loss)
            cellstate_models.append(model)

        last_delta_losses_mean = np.array(
            [
                float(torch.mean(losses[-delta_loss_batch:]))
                for losses in cellstate_losses
            ]
        )

        best_model_index = int(np.argmin(last_delta_losses_mean))
        self._state_ast = cellstate_models[best_model_index]

        self._state_ast.fit(
            max_epochs=max_epochs,
            learning_rate=learning_rate,
            batch_size=batch_size,
            delta_loss=delta_loss,
            delta_loss_batch=delta_loss_batch,
            msg=" (final)",
        )

        # Warns the user if the model has not converged
        if not self._state_ast.is_converged():
            msg = (
                "Maximum epochs reached. More iteration may be needed to"
                + " complete the training."
            )
            warnings.warn(msg)

        self._state_run_info = {
            "max_epochs": max_epochs,
            "learning_rate": learning_rate,
            "batch_size": batch_size,
            "delta_loss": delta_loss,
            "n_init": n_init,
            "n_init_epochs": n_init_epochs,
            "delta_loss_batch": delta_loss_batch,
            "const": const,
            "dropout_rate": dropout_rate,
            "batch_norm": batch_norm,
        }

    def save_models(self, hdf5_name: str = "astir_summary.hdf5") -> None:
        """Save the summary of this model to an `hdf5` file.

        :param hdf5_name: name of the output `hdf5` file, default to "astir_summary.hdf5"
        :raises Exception: raised when this function is called before the model is trained.
        """
        if self._type_ast is None and self._state_ast is None:
            raise Exception("No model has been trained")
        with h5py.File(hdf5_name, "w") as f:
            if self._type_ast is not None:
                type_grp = f.create_group("celltype_model")

                # Storing losses
                loss_grp = type_grp.create_group("losses")
                loss_grp["losses"] = self.get_type_losses().cpu().numpy()

                # Storing parameters
                param_grp = type_grp.create_group("parameters")
                dic = list(self._type_ast.get_variables().items()) + list(
                    self._type_ast.get_data().items()
                )
                for key, val in dic:
                    param_grp[key] = val.detach().cpu().numpy()

                # Storing Recognition Network
                recog_grp = type_grp.create_group("recog_net")
                for name, param in self._type_ast.get_recognet().named_parameters():
                    if param.requires_grad:
                        recog_grp[name] = param.detach().cpu().numpy()

                # Storing fit_type argument information
                info_grp = type_grp.create_group("run_info")
                for key, val in self._type_run_info.items():
                    info_grp[key] = val

            if self._state_ast is not None:
                state_grp = f.create_group("cellstate_model")

                # Storing losses
                loss_grp = state_grp.create_group("losses")
                loss_grp["losses"] = self.get_state_losses()

                # Storing parameters
                param_grp = state_grp.create_group("parameters")
                dic = list(self._state_ast.get_variables().items()) + list(
                    self._state_ast.get_data().items()
                )

                # Storing Recognition Network
                recog_grp = state_grp.create_group("recog_net")
                for name, param in self._state_ast.get_recognet().named_parameters():
                    if param.requires_grad:
                        recog_grp[name] = param.detach().cpu().numpy()

                for key, val in dic:
                    param_grp[key] = val.detach().cpu().numpy()

                # Storing fit_state method arguments
                info_grp = state_grp.create_group("run_info")
                for key, val in self._state_run_info.items():
                    info_grp[key] = val

        if self._type_ast is not None:
            self._type_ast.get_assignment().to_hdf(
                hdf5_name, "/celltype_model/celltype_assignments"
            )
        if self._state_ast is not None:
            self._state_ast.get_assignment().to_hdf(
                hdf5_name, "/cellstate_model/cellstate_assignments"
            )

    def load_model(self, hdf5_name: str) -> None:
        """Load model from hdf5 file

        :param hdf5_name: the full path to file
        """
        has_type = False
        has_state = False
        with h5py.File(hdf5_name, "r") as f:
            if "celltype_model" in f.keys():
                run_info = f["celltype_model/run_info"]
                self._type_run_info = {
                    "learning_rate": float(np.array(run_info["learning_rate"])),
                    "n_init_epochs": int(np.array(run_info["n_init_epochs"])),
                    "batch_size": float(np.array(run_info["batch_size"])),
                    "n_init": int(np.array(run_info["n_init"])),
                    "delta_loss": float(np.array(run_info["delta_loss"])),
                    "max_epochs": int(np.array(run_info["max_epochs"])),
                }
                has_type = True
            if "cellstate_model" in f.keys():
                run_info = f["cellstate_model/run_info"]
                const = int(np.array(run_info["const"]))
                dropout_rate = float(np.array(run_info["dropout_rate"]))
                batch_norm = bool(np.array(run_info["batch_norm"]))
                self._state_run_info = {
                    "learning_rate": float(np.array(run_info["learning_rate"])),
                    "n_init_epochs": int(np.array(run_info["n_init_epochs"])),
                    "batch_size": float(np.array(run_info["batch_size"])),
                    "n_init": int(np.array(run_info["n_init"])),
                    "delta_loss": float(np.array(run_info["delta_loss"])),
                    "max_epochs": int(np.array(run_info["max_epochs"])),
                    "delta_loss_batch": int(np.array(run_info["delta_loss_batch"])),
                    "const": const,
                    "dropout_rate": dropout_rate,
                    "batch_norm": batch_norm,
                }
                has_state = True
        if has_type:
            self._type_ast = CellTypeModel(
                dset=self._type_dset,
                dtype=self._dtype,
                random_seed=np.random.randint(9999),
            )
            self._type_ast.load_hdf5(hdf5_name)
        if has_state:
            self._state_ast = CellStateModel(
                dset=self._type_dset,
                const=const,
                dropout_rate=dropout_rate,
                batch_norm=batch_norm,
                random_seed=np.random.randint(9999),
                dtype=self._dtype,
                device=self._device,
            )
            self._state_ast.load_hdf5(hdf5_name)

    def get_type_dataset(self) -> SCDataset:
        """Get the `SCDataset` for cell type training.

        :return: `self._type_dset`
        """
        if self._type_dset is None:
            raise Exception("the type dataset is not provided")
        return self._type_dset

    def get_state_dataset(self) -> SCDataset:
        """Get the `SCDataset` for cell state training.

        :return: `self._state_dset`
        """
        if self._state_dset is None:
            raise Exception("the state dataset is not provided")
        return self._state_dset

    def get_type_model(self) -> CellTypeModel:
        """Get the trained `CellTypeModel`.

        :raises Exception: raised when this function is called before the model is trained.
        :return: `self._type_ast`
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        return self._type_ast

    def get_state_model(self) -> CellStateModel:
        """Get the trained `CellStateModel`.

        :raises Exception: raised when this function is celled before the model is trained.
        :return: `self._state_ast`
        """
        if self._state_ast is None:
            raise Exception("The state model has not been trained yet")
        return self._state_ast

    def get_type_run_info(self) -> Dict[str, Union[int, float]]:
        """Get the run information (i.e. `max_epochs`, `learning_rate`,
            `batch_size`, `delta_loss`, `n_init`, `n_init_epochs`) of the cell type training.

        :raises Exception: raised when this function is celled before the model is trained.
        :return: `self._type_run_info`
        """
        if self._type_run_info == {}:
            raise Exception("The type model has not been trained yet")
        return self._type_run_info

    def get_state_run_info(self) -> Dict[str, Union[int, float]]:
        """Get the run information (i.e. `max_epochs`, `learning_rate`, `batch_size`,
            `delta_loss`, `n_init`, `n_init_epochs`, `delta_loss_batch`) of the cell state training.

        :raises Exception: raised when this function is celled before the model is trained.
        :return: `self._state_run_info`
        """
        if self._state_run_info == {}:
            raise Exception("The state model has not been trained yet")
        return self._state_run_info

    def get_celltype_probabilities(self) -> pd.DataFrame:
        """Get the cell assignment probability.

        :return: `self.assignments`
        :rtype: pd.DataFrame
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        return self._type_ast.get_assignment()

    def get_cellstates(self) -> pd.DataFrame:
        """Get cell state activations. It returns the rescaled activations,
        values between 0 and 1

        :return: state assignments
        :rtype: pd.DataFrame
        """
        if self._state_ast is None:
            raise Exception("The state model has not been trained yet")

        assign = self._state_ast.get_assignment()
        assign_min = assign.min(axis=0)
        assign_max = assign.max(axis=0)

        assign_rescale = (assign - assign_min) / (assign_max - assign_min)

        return assign_rescale

    def get_celltypes(
        self, threshold: float = 0.7, assignment_type: str = "threshold"
    ) -> pd.DataFrame:
        """Get the most likely cell type

        A cell is assigned to a cell type if the probability is greater than threshold.
        If no cell types have a probability higher than threshold, then "Unknown" is returned

        :param threshold: the probability threshold above which a cell is assigned to a cell type
        :param assignment_type: See
            :meth:`astir.CellTypeModel.get_celltypes` for full documentation
        :return: a data frame with most likely cell types for each
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        return self._type_ast.get_celltypes(threshold, assignment_type=assignment_type)

    def predict_celltypes(self, dset: pd.DataFrame = None) -> pd.DataFrame:
        """Predict the probabilities of different cell type assignments.

        :param dset: the single cell protein expression dataset to predict, defaults to None
        :type dset: pd.DataFrame, optional
        :raises Exception: when the type model is not trained when this function is called
        :return: the probabilities of different cell type assignments
        :rtype: pd.DataFrame
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        if not self._type_ast.is_converged():
            msg = "The state model has not been trained for enough epochs yet"
            warnings.warn(msg)
        if dset is None:
            dset = self.get_type_dataset()

        type_assignments = self._type_ast.predict(dset)
        type_assignments.columns = dset.get_classes() + ["Other"]
        type_assignments.index = dset.get_cell_names()
        return type_assignments

    def predict_cellstates(self, dset: SCDataset = None) -> pd.DataFrame:
        """Get the prediction cell state activations on a dataset on an
        existing model

        :param new_dset: the dataset to predict cell state activations, default to None

        :return: the prediction of cell state activations
        """
        if self._state_ast is None:
            raise Exception("The state model has not been trained yet")

        if not self._state_ast.is_converged():
            msg = "The state model has not been trained for enough epochs yet"
            warnings.warn(msg)

        if self._state_dset is None:
            raise NotClassifiableError(
                "Dataset or marker for cell state classification is not provided"
            )

        g = self._state_ast.get_final_mu_z(dset).detach().cpu().numpy()

        state_assignments = pd.DataFrame(g)

        assign_min = state_assignments.min(axis=0)
        assign_max = state_assignments.max(axis=0)

        assign_rescale = (state_assignments - assign_min) / (assign_max - assign_min)

        assign_rescale.columns = self._state_dset.get_classes()
        assign_rescale.index = self._state_dset.get_cell_names()

        return assign_rescale

    def assign_celltype_hierarchy(self, depth: int = 1) -> pd.DataFrame:
        """Get cell type assignment at a specified higher hierarchy according to the hierarchy provided
            in the dictionary.

        :param depth: the depth of hierarchy to assign probability to, defaults to 1
        :type depth: int, optional
        :raises Exception: raised when the dictionary for hierarchical structure is not provided
            or the model hasn't been trained.
        :return: probability assignment of cell type at a superstructure
        :rtype: pd.DataFrame
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        if self._hierarchy_dict is None:
            raise Exception("The dictionary for hierarchical structure is not provided")
        prob = self.get_celltype_probabilities()
        hier_df = pd.DataFrame()
        self._assign_celltype_hierarchy_helper1(
            hier_df, prob, self._hierarchy_dict, depth
        )
        return hier_df

    def _assign_celltype_hierarchy_helper1(
        self,
        hier_df: Union[pd.DataFrame, list],
        prob: pd.DataFrame,
        dic: dict,
        depth: int = 1,
    ) -> None:
        """Helper for `assign_celltype_hierarchy`, calculates summed probabilities recursively."""
        if depth == 1:
            for key, val in dic.items():
                hier_df[key] = self._assign_celltype_hierarchy_helper2(prob, val)
        else:
            for key, cells in dic.items():
                if isinstance(cells, list):
                    for cell in cells:
                        hier_df[cell] = prob[cell]
                else:
                    self._assign_celltype_hierarchy_helper1(
                        hier_df, prob, cells, depth - 1
                    )

    def _assign_celltype_hierarchy_helper2(
        self, prob: pd.DataFrame, dic: dict
    ) -> pd.Series:
        """Helper for `assign_celltype_hierarchy`, calculates summed probability for
        cells under the given dict.
        """
        if isinstance(dic, list):
            return prob[dic].sum(axis=1)
        temp_df = pd.DataFrame()
        for key, cells in dic.items():
            temp_df[key] = self._assign_celltype_hierarchy_helper2(prob, cells)
        return temp_df.sum(axis=1)

    def type_clustermap(
        self,
        plot_name: str = "celltype_protein_cluster.png",
        threshold: float = 0.7,
        figsize: Tuple[float, float] = (7, 5),
        prob_assign: Optional[pd.DataFrame] = None,
    ) -> None:
        """Save the heatmap of protein content in cells with cell types labeled.

        :param plot_name: name of the plot, extension(e.g. .png or .jpg) is needed, defaults to "celltype_protein_cluster.png"
        :type plot_name: str, optional
        :param threshold: the probability threshold above which a cell is assigned to a cell type, defaults to 0.7
        :type threshold: float, optional
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        self._type_ast.plot_clustermap(plot_name, threshold, figsize, prob_assign)

    def get_hierarchy_dict(self) -> Dict[str, List[str]]:
        """Get the dictionary for cell type hierarchical structure.

        :return: `self._hierarchy_dict`
        :rtype: Dict[str, List[str]]
        """
        if self._hierarchy_dict is None:
            raise Exception("the hierarchical marker is not provided")
        return self._hierarchy_dict

    def get_type_losses(self) -> np.array:
        """Get the final losses of the type model.

        :return: `self.losses`
        :rtype: np.array
        """
        if self._type_ast is None:
            raise Exception("The type model has not been trained yet")
        return self._type_ast.get_losses()

    def get_state_losses(self) -> np.array:
        """Getter for losses

        :return: a numpy array of losses for each training iteration the\
            model runs
        :rtype: np.array
        """
        if self._state_ast is None:
            raise Exception("The state model has not been trained yet")
        return self._state_ast.get_losses()

    def type_to_csv(
        self,
        output_csv: str,
        threshold: float = 0.7,
        assignment_type: str = "threshold",
    ) -> None:
        """Save the cell type assignemnt to a `csv` file.

        :param output_csv: name for the output .csv file
        :param assignment_type: See
            :meth:`astir.CellTypeModel.get_celltypes` for full documentation
        :type output_csv: str
        """
        self.get_celltypes(threshold, assignment_type=assignment_type).to_csv(
            output_csv
        )

    def state_to_csv(self, output_csv: str) -> None:
        """Writes state assignment output from training state model in csv
        file

        :param output_csv: path to output csv
        :type output_csv: str, required
        """
        self.get_cellstates().to_csv(output_csv)

    def __str__(self) -> str:
        """String representation of Astir object"""
        msg = "Astir object"
        l = 0
        if self._type_dset is not None:
            msg += ", " + str(self._type_dset.get_n_classes()) + " cell types"
            l = len(self._type_dset)
        if self._state_dset is not None:
            msg += ", " + str(self._state_dset.get_n_classes()) + " cell states"
            l = len(self._state_dset)
        if l != 0:
            msg += ", " + str(l) + " cells"
        return msg

    def diagnostics_celltype(
        self, threshold: float = 0.7, alpha: float = 0.01
    ) -> pd.DataFrame:
        """Run diagnostics on cell type assignments

        This performs a basic test that cell types express their markers at
        higher levels than in other cell types. This function performs the following steps:

        1. Iterates through every cell type and every marker for that cell type
        2. Given a cell type *c* and marker *g*, find the set of cell
            types *D* that don't have *g* as a marker
        3. For each cell type *d* in *D*, perform a t-test between the
            expression of marker *g* in *c* vs *d*
        4. If *g* is not expressed significantly higher (at significance
            *alpha*), output a diagnostic explaining this for further investigation.

        :param threshold: The threshold at which cell types are assigned (see `get_celltypes`)
        :param alpha: The significance threshold for t-tests for determining over-expression
        :return: Either a :class:`pd.DataFrame` listing cell types whose markers
            aren't expressed signficantly higher.
        """
        celltypes = list(self.get_celltypes(threshold=threshold)["cell_type"])
        return self.get_type_model().diagnostics(celltypes, alpha=alpha)

    def diagnostics_cellstate(self) -> pd.DataFrame:
        """Run diagnostics on cell state assignments

        This performs a basic test by comparing the correlation values
        between all marker genes and all non marker genes. It detects where the
        non marker gene has higher correlation values than the smallest
        correlation values of marker genes.

        1. Get correlations between all cell states and proteins
        2. For each cell state *c*, get the smallest correlation with marker *g*
        3. For each cell state *c* and its non marker *g*, find any correlation that is
            bigger than those smallest correlation for *c*.
        4. Any *c* and *g* pairs found in step 3 will be included in the output of
            `Astir.diagnostics_cellstate()`, including an explanation.

        :return: diagnostics
        """
        return self.get_state_model().diagnostics()

    def normalize(self, percentile_lower: int = 1, percentile_upper: int = 99) -> None:
        """Normalize the expression data

        This performs a two-step normalization:
        1. A `log(1+x)` transformation to the data
        2. Winsorizes to (`percentile_lower`, `percentile_upper`)

        :param percentile_lower: Lower percentile for winsorization
        :param percentile_upper: Upper percentile for winsorization
        """
        self.get_type_dataset().normalize(percentile_lower, percentile_upper)
        self.get_state_dataset().normalize(percentile_lower, percentile_upper)


class NotClassifiableError(RuntimeError):
    """Raised when the input data is not classifiable."""

    pass


# KC removed
# :param assignments: cell type assignment probabilities
# :param losses: losses after optimization
# :param type_dict: dictionary mapping cell type
#     to the corresponding genes
# :param state_dict: dictionary mapping cell state
#     to the corresponding genes
# :param cell_types: list of all cell types from marker
# :param mtype_genes: list of all cell type genes from marker
# :param mstate_genes: list of all cell state genes from marker
# :param N: number of rows of data
# :param G: number of cell type genes
# :param C: number of cell types
# :param initializations: initialization parameters
# :param data: parameters that is not to be optimized
# :param variables: parameters that is to be optimized



from .astir import Astir
from .data.data_readers import (
    from_anndata_yaml,
    from_csv_dir_yaml,
    from_csv_yaml,
    from_loompy_yaml,
)
from .data.scdataset import SCDataset
from .models.cellstate import CellStateModel
from .models.celltype import CellTypeModel

__all__ = ["Astir", "NotClassifiableError"]



"""
State Recognition Neural Network Model
"""

import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


# The recognition net
class StateRecognitionNet(nn.Module):
    """State Recognition Neural Network to get mean of z and standard
    deviation of z. The neural network architecture looks like this: G ->
    const * C -> const * C -> G (for mu) or -> G (for std). With batch
    normal layers after each activation output layers and dropout
    activation units

    :param C: the number of pathways
    :param G: the number of proteins
    :param const: the size of the hidden layers are const times proportional
        to C, defaults to 2
    :param dropout_rate: the dropout rate, defaults to 0
    :param batch_norm: apply batch normal layers if True, defaults to False
    """

    def __init__(
        self,
        C: int,
        G: int,
        const: int = 2,
        dropout_rate: float = 0,
        batch_norm: bool = False,
    ) -> None:
        super(StateRecognitionNet, self).__init__()
        self.batch_norm = batch_norm

        hidden_layer_size = math.ceil(const * C)
        # First hidden layer
        self.linear1 = nn.Linear(G, hidden_layer_size).float()
        self.dropout1 = nn.Dropout(dropout_rate)

        # Second hidden layer
        self.linear2 = nn.Linear(hidden_layer_size, hidden_layer_size).float()
        self.dropout2 = nn.Dropout(dropout_rate)

        # Output layer for mu
        self.linear3_mu = nn.Linear(hidden_layer_size, C).float()
        self.dropout_mu = nn.Dropout(dropout_rate)

        # Output layer for std
        self.linear3_std = nn.Linear(hidden_layer_size, C).float()
        self.dropout_std = nn.Dropout(dropout_rate)

        # Batch normal layers
        if self.batch_norm:
            self.bn1 = nn.BatchNorm1d(num_features=hidden_layer_size).float()
            self.bn2 = nn.BatchNorm1d(num_features=hidden_layer_size).float()
            self.bn_out_mu = nn.BatchNorm1d(num_features=C).float()
            self.bn_out_std = nn.BatchNorm1d(num_features=C).float()

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """One forward pass of the StateRecognitionNet

        :param x: the input to the recognition network model
        :return: the value from the output layer of the network
        """
        # Input --linear1--> Hidden1
        x = self.linear1(x)
        if self.batch_norm:
            x = self.bn1(x)
        x = F.relu(x)
        x = self.dropout1(x)

        # Hidden1 --linear2--> Hidden2
        x = self.linear2(x)
        if self.batch_norm:
            x = self.bn2(x)
        x = F.relu(x)
        x = self.dropout2(x)

        # Hidden2 --linear3_mu--> mu
        mu_z = self.linear3_mu(x)
        if self.batch_norm:
            mu_z = self.bn_out_mu(mu_z)
        mu_z = self.dropout_mu(mu_z)

        # Hidden2 --linear3_std--> std
        std_z = self.linear3_std(x)
        if self.batch_norm:
            std_z = self.bn_out_std(std_z)
        std_z = self.dropout_std(std_z)

        return mu_z, std_z



""" 
Cell Type Model
"""

import re
import warnings
from collections import OrderedDict
from typing import Dict, Generator, List, Optional, Tuple, Union

import h5py
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Dirichlet
from scipy import stats
from sklearn.preprocessing import StandardScaler
from torch.autograd import Variable
from torch.distributions import (
    LowRankMultivariateNormal,
    MultivariateNormal,
    Normal,
    StudentT,
)
from torch.utils.data import DataLoader, Dataset
from tqdm import trange

from astir.data import SCDataset

from .abstract import AstirModel
from .celltype_recognet import TypeRecognitionNet


class CellTypeModel(AstirModel):
    """Class to perform statistical inference to assign cells to cell types.

    :param dset: the input gene expression dataframe
    :param random_seed: the random seed for parameter initialization, defaults to 1234
    :param dtype: the data type of parameters, should be the same as `dset`, defaults to
        torch.float64
    """

    def __init__(
        self,
        dset: Optional[SCDataset] = None,
        random_seed: int = 1234,
        dtype: torch.dtype = torch.float64,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        super().__init__(dset, random_seed, dtype, device)

        if dset is not None:
            self._param_init()

    def _param_init(self) -> None:
        """Initializes parameters and design matrices."""
        if self._dset is None:
            raise Exception("the dataset is not provided")
        G = self._dset.get_n_features()
        C = self._dset.get_n_classes()

        self._recog = TypeRecognitionNet(
            self._dset.get_n_classes(), self._dset.get_n_features()
        ).to(self._device, dtype=self._dtype)

        # Establish data
        self._data: Dict[str, torch.Tensor] = {
            # "log_alpha": torch.log(torch.ones(C + 1, dtype=self._dtype) / (C + 1)).to(
            #     self._device
            # ),
            "rho": self._dset.get_marker_mat().to(self._device),
        }

        self._alpha_prior = Dirichlet(
            torch.ones(C + 1, dtype=self._dtype).to(self._device) * (C + 1)
        )

        # Initialize mu, log_delta
        delta_init_mean = torch.log(
            torch.log(torch.tensor(3.0, dtype=self._dtype))
        )  # the log of the log of this is the multiplier
        t = torch.distributions.Normal(
            # delta_init_mean.clone().detach().to(self._dtype),
            torch.tensor(0, dtype=self._dtype),
            torch.tensor(0.1, dtype=self._dtype),
        )
        log_delta_init = t.sample((G, C + 1))

        mu_init = torch.log(
            torch.tensor(self._dset.get_mu_init(), dtype=self._dtype)
        ).to(self._device)
        # mu_init = torch.log(self._dset.get_mu()).to(self._device)
        # mu_init = mu_init - (
        #     self._data["rho"] * torch.exp(log_delta_init).to(self._device)
        # ).mean(1)

        mu_init = mu_init.reshape(-1, 1)

        # Create initialization dictionary
        initializations = {
            "mu": mu_init,
            "log_sigma": torch.log(self._dset.get_sigma()).to(self._device),
            "log_delta": log_delta_init,
            "p": torch.zeros((G, C + 1), dtype=self._dtype, device=self._device),
            "alpha_logits": torch.ones(C + 1, dtype=self._dtype, device=self._device),
        }
        P = self._dset.get_design().shape[1]
        # Add additional columns of mu for anything in the design matrix
        initializations["mu"] = torch.cat(
            [
                initializations["mu"],
                torch.zeros((G, P - 1), dtype=self._dtype, device=self._device),
            ],
            1,
        )
        # Create trainable variables
        self._variables: Dict[str, torch.Tensor] = {}
        for (n, v) in initializations.items():
            self._variables[n] = Variable(v.clone()).to(self._device)
            self._variables[n].requires_grad = True

    def load_hdf5(self, hdf5_name: str) -> None:
        """Initializes Cell Type Model from a hdf5 file type

        :param hdf5_name: file path
        """
        self._assignment = pd.read_hdf(hdf5_name, "celltype_model/celltype_assignments")
        with h5py.File(hdf5_name, "r") as f:
            grp = f["celltype_model"]
            param = grp["parameters"]
            self._variables = {
                "mu": torch.tensor(np.array(param["mu"])),
                "log_sigma": torch.tensor(np.array(param["log_sigma"])),
                "log_delta": torch.tensor(np.array(param["log_delta"])),
                "p": torch.tensor(np.array(param["p"])),
                "alpha_logits": torch.tensor(np.array(param["alpha_logits"])),
            }
            self._data = {
                "rho": torch.tensor(np.array(param["rho"])),
            }
            self._losses = torch.tensor(np.array(grp["losses"]["losses"]))

            rec = grp["recog_net"]
            hidden1_W = torch.tensor(np.array(rec["hidden_1.weight"]))
            hidden2_W = torch.tensor(np.array(rec["hidden_2.weight"]))
            state_dict = {
                "hidden_1.weight": hidden1_W,
                "hidden_1.bias": torch.tensor(np.array(rec["hidden_1.bias"])),
                "hidden_2.weight": hidden2_W,
                "hidden_2.bias": torch.tensor(np.array(rec["hidden_2.bias"])),
            }
            state_dict = OrderedDict(state_dict)
            self._recog = TypeRecognitionNet(
                hidden2_W.shape[0] - 1, hidden1_W.shape[1], hidden1_W.shape[0]
            ).to(device=self._device, dtype=self._dtype)
            self._recog.load_state_dict(state_dict)
            self._recog.eval()

    def _forward(
        self, Y: torch.Tensor, X: torch.Tensor, design: torch.Tensor
    ) -> torch.Tensor:
        """One forward pass.

        :param Y: a sample from the dataset
        :param X: normalized sample data
        :param design: the corresponding row of design matrix
        :return: the cost (elbo) of the current pass
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        G = self._dset.get_n_features()
        C = self._dset.get_n_classes()
        N = Y.shape[0]

        Y_spread = Y.view(N, 1, G).repeat(1, C + 1, 1)

        delta_tilde = torch.exp(self._variables["log_delta"])
        mean = delta_tilde * self._data["rho"]
        mean2 = torch.mm(design, self._variables["mu"].T)  ## N x P * P x G
        mean2 = mean2.view(-1, G, 1).repeat(1, 1, C + 1)
        mean = mean + mean2

        # now do the variance modelling
        p = torch.sigmoid(self._variables["p"])

        sigma = torch.exp(self._variables["log_sigma"])
        v1 = (self._data["rho"] * p).T * sigma
        v2 = torch.pow(sigma, 2) * (1 - torch.pow(self._data["rho"] * p, 2)).T

        v1 = v1.view(1, C + 1, G, 1).repeat(N, 1, 1, 1)  # extra 1 is the "rank"
        v2 = v2.view(1, C + 1, G).repeat(N, 1, 1) + 1e-6

        dist = LowRankMultivariateNormal(
            loc=torch.exp(mean).permute(0, 2, 1), cov_factor=v1, cov_diag=v2
        )

        log_p_y_on_c = dist.log_prob(Y_spread)

        gamma, log_gamma = self._recog.forward(X)
        log_alpha = F.log_softmax(self._variables["alpha_logits"], dim=0)
        alpha = F.softmax(self._variables["alpha_logits"], dim=0)
        mix_prior = self._alpha_prior.log_prob(alpha)

        elbo = (gamma * (log_p_y_on_c + log_alpha - log_gamma)).sum() + mix_prior

        return -elbo

    def fit(
        self,
        max_epochs: int = 50,
        learning_rate: float = 1e-3,
        batch_size: int = 128,
        delta_loss: float = 1e-3,
        delta_loss_batch: int = 10,
        msg: str = "",
    ) -> None:
        """Runs train loops until the convergence reaches delta_loss for
        delta_loss_batch sizes or for max_epochs number of times

        :param max_epochs: number of train loop iterations, defaults to 50
        :param learning_rate: the learning rate, defaults to 0.01
        :param batch_size: the batch size, defaults to 128
        :param delta_loss: stops iteration once the loss rate reaches
            delta_loss, defaults to 0.001
        :param delta_loss_batch: the batch size to consider delta loss,
            defaults to 10
        :param msg: iterator bar message, defaults to empty string
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        # Make dataloader
        dataloader = DataLoader(
            self._dset, batch_size=min(batch_size, len(self._dset)), shuffle=True
        )

        # Run training loop
        losses: List[torch.Tensor] = []
        per = torch.tensor(1)

        # Construct optimizer
        opt_params = list(self._variables.values()) + list(self._recog.parameters())

        optimizer = torch.optim.Adam(opt_params, lr=learning_rate)

        _, exprs_X, _ = self._dset[:]  # calls dset.get_item

        iterator = trange(
            max_epochs,
            desc="training restart" + msg,
            unit="epochs",
            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{rate_fmt}{postfix}]",
        )
        for ep in iterator:
            # for ep in range(max_epochs):
            L = None
            loss = torch.tensor(0.0, dtype=self._dtype)
            for batch in dataloader:
                Y, X, design = batch
                optimizer.zero_grad()
                L = self._forward(Y, X, design)
                L.backward()
                optimizer.step()
                with torch.no_grad():
                    loss = loss + L
            if len(losses) > 0:
                per = abs((loss - losses[-1]) / losses[-1])
            losses.append(loss)
            iterator.set_postfix_str("current loss: " + str(round(float(loss), 1)))

            if per <= delta_loss:
                self._is_converged = True
                iterator.close()
                break

        # Save output
        self._assignment = pd.DataFrame(
            self._recog.forward(exprs_X)[0].detach().cpu().numpy()
        )
        self._assignment.columns = self._dset.get_classes() + ["Other"]
        self._assignment.index = self._dset.get_cell_names()

        if self._losses.shape[0] == 0:
            self._losses = torch.tensor(losses)
        else:
            self._losses = torch.cat(
                (self._losses.view(self._losses.shape[0]), torch.tensor(losses)), dim=0
            )

    def predict(self, new_dset: pd.DataFrame) -> np.array:
        """Feed `new_dset` to the recognition net to get a prediction.

        :param new_dset: the dataset to be predicted
        :return: the resulting cell type assignment
        """
        _, exprs_X, _ = new_dset[:]
        g = pd.DataFrame(self._recog.forward(exprs_X)[0].detach().cpu().numpy())
        return g

    def get_recognet(self) -> TypeRecognitionNet:
        """Getter for the recognition net.

        :return: the trained recognition net
        """
        return self._recog

    def _most_likely_celltype(
        self,
        row: pd.DataFrame,
        threshold: float,
        cell_types: List[str],
        assignment_type: str,
    ) -> str:
        """Given a row of the assignment matrix, return the most likely cell type

        :param row: the row of cell assignment matrix to be evaluated
        :param threshold: the higher bound of the maximun probability to classify a cell as `Unknown`
        :param cell_types: the names of cell types, in the same order as the features of the row
        :param assignment_type: See
        :meth:`astir.CellTypeModel.get_celltypes` for full documentation
        :return: the most likely cell type of this cell
        """
        row = row.values
        max_prob = np.max(row)

        if assignment_type == "threshold":
            if max_prob < threshold:
                return "Unknown"
        elif assignment_type == "max":
            if sum(row == max_prob) > 1:
                return "Unknown"

        return cell_types[np.argmax(row)]

    def get_celltypes(
        self,
        threshold: float = 0.7,
        assignment_type: str = "threshold",
        prob_assign: Optional[pd.DataFrame] = None,
    ) -> pd.DataFrame:
        """Get the most likely cell types. A cell is assigned to a cell type
        if the probability is greater than threshold.
        If no cell types have a probability higher than threshold,
        then "Unknown" is returned.

        :param assignment_type: either 'threshold' or 'max'. If threshold,
            type assignment is based on whether the probability threshold is
            above prob_assignment. If 'max', type assignment is based on the max
            probability value or "unknown" if there are multiple max
            probabilities. Defaults to 'threshold'.
        :param threshold: the probability threshold above which a cell is
            assigned to a cell type, defaults to 0.7
        :return: a data frame with most likely cell types for each
        """
        if prob_assign is None:
            type_probability = self.get_assignment()
        else:
            type_probability = prob_assign

        if assignment_type != "threshold" and assignment_type != "max":
            warnings.warn(
                "Wrong assignment type. Defaults the assignment " "type to threshold."
            )
            assignment_type = "threshold"

        if assignment_type == "max" and prob_assign is not None:
            warnings.warn(
                "Assignment type is 'max' but probability "
                "threshold value was passed in. Probability "
                "threshold value will be ignored."
            )

        cell_types = list(type_probability.columns)

        cell_type_assignments = type_probability.apply(
            self._most_likely_celltype,
            axis=1,
            assignment_type=assignment_type,
            threshold=threshold,
            cell_types=cell_types,
        )
        cell_type_assignments = pd.DataFrame(cell_type_assignments)
        cell_type_assignments.columns = ["cell_type"]

        return cell_type_assignments

    def _compare_marker_between_types(
        self,
        curr_type: str,
        celltype_to_compare: str,
        marker: str,
        cell_types: List[str],
        alpha: float = 0.05,
    ) -> Optional[dict]:
        """For two cell types and a protein, ensure marker
        is expressed at higher level for curr_type than celltype_to_compare

        :param curr_type: the cell type to assess
        :param celltype_to_compare: all the cell types that shouldn't highly express this marker
        :param marker: the marker protein for curr_type
        :param cell_types: list of cell types assigned for cells
        :param alpha:
        :return:
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        current_marker_ind = np.array(self._dset.get_features()) == marker

        cells_x = np.array(cell_types) == curr_type
        cells_y = np.array(cell_types) == celltype_to_compare

        # x - cells whose cell types' marker protein is marker
        # y - cells whose cell types' marker protein is not marker
        x = self._dset.get_exprs().detach().cpu().numpy()[cells_x, current_marker_ind]
        y = self._dset.get_exprs().detach().cpu().numpy()[cells_y, current_marker_ind]

        stat = np.NaN
        pval = np.Inf
        note: Optional[str] = "Only 1 cell in a type: comparison not possible"

        if len(x) > 1 and len(y) > 1:
            tt = stats.ttest_ind(x, y)
            stat = tt.statistic
            pval = tt.pvalue
            note = None

        if not (stat > 0 and pval < alpha):
            rdict = {
                "current_marker": marker,
                "curr_type": curr_type,
                "celltype_to_compare": celltype_to_compare,
                "mean_A": x.mean(),
                "mean_Y": y.mean(),
                "p-val": pval,
                "note": note,
            }

            return rdict

        return None

    def plot_clustermap(
        self,
        plot_name: str = "celltype_protein_cluster.png",
        threshold: float = 0.7,
        figsize: Tuple[float, float] = (7.0, 5.0),
        prob_assign: Optional[pd.DataFrame] = None,
    ) -> None:
        """Save the heatmap of protein content in cells with cell types labeled.

        :param plot_name: name of the plot, extension(e.g. .png or .jpg) is needed, defaults to "celltype_protein_cluster.png"
        :param threshold: the probability threshold above which a cell is assigned to a cell type, defaults to 0.7
        :param figsize: the size of the figure, defaults to (7.0, 5.0)
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        expr_df = self._dset.get_exprs_df()
        scaler = StandardScaler()
        for feature in expr_df.columns:
            expr_df[feature] = scaler.fit_transform(
                expr_df[feature].values.reshape((expr_df[feature].shape[0], 1))
            )

        expr_df["cell_type"] = self.get_celltypes(
            threshold=threshold, prob_assign=prob_assign
        )
        expr_df = expr_df.sort_values(by=["cell_type"])
        types = expr_df.pop("cell_type")
        types_uni = types.unique()

        lut = dict(zip(types_uni, sns.color_palette("BrBG", len(types_uni))))
        col_colors = pd.DataFrame(types.map(lut))
        cm = sns.clustermap(
            expr_df.T,
            xticklabels=False,
            cmap="vlag",
            col_cluster=False,
            col_colors=col_colors,
            figsize=figsize,
        )

        for t in types_uni:
            cm.ax_col_dendrogram.bar(0, 0, color=lut[t], label=t, linewidth=0)
        cm.ax_col_dendrogram.legend(
            title="Cell Types", loc="center", ncol=3, bbox_to_anchor=(0.8, 0.8)
        )
        cm.savefig(plot_name, dpi=150)

    def diagnostics(self, cell_type_assignments: list, alpha: float) -> pd.DataFrame:
        """Run diagnostics on cell type assignments

        See :meth:`astir.Astir.diagnostics_celltype` for full documentation
        """
        if self._dset is None:
            raise Exception("the dataset is not provided")
        problems = []

        # Want to construct a data frame that models rho with
        # cell type names on the columns and feature names on the rows
        g_df = pd.DataFrame(self._data["rho"].detach().cpu().numpy())
        g_df.columns = self._dset.get_classes() + ["Other"]
        g_df.index = self._dset.get_features()

        for curr_type in self._dset.get_classes():
            if not curr_type in cell_type_assignments:
                continue

            current_markers = g_df.index[g_df[curr_type] == 1]

            for current_marker in current_markers:
                # find all the cell types that shouldn't highly express this marker
                celltypes_to_compare = g_df.columns[g_df.loc[current_marker] == 0]

                for celltype_to_compare in celltypes_to_compare:
                    if not celltype_to_compare in cell_type_assignments:
                        continue

                    is_problem = self._compare_marker_between_types(
                        curr_type,
                        celltype_to_compare,
                        current_marker,
                        cell_type_assignments,
                        alpha,
                    )

                    if is_problem is not None:
                        problems.append(is_problem)

        col_names = [
            "feature",
            "should be expressed higher in",
            "than",
            "mean cell type 1",
            "mean cell type 2",
            "p-value",
            "note",
        ]
        df_issues = None
        if len(problems) > 0:
            df_issues = pd.DataFrame(problems)
            df_issues.columns = col_names
        else:
            df_issues = pd.DataFrame(columns=col_names)

        return df_issues



import os
import warnings
from unittest import TestCase

import pandas as pd
import torch
import yaml

from astir.data import SCDataset, from_csv_yaml
from astir.models import CellTypeModel


class TestCellTypeModel(TestCase):
    """Unittest class for CellTypeModel class

    This class assumes that all data initializating functions in Astir class
    are working
    """

    def __init__(self, *args, **kwargs):
        super(TestCellTypeModel, self).__init__(*args, **kwargs)

        self.expr_csv_file = os.path.join(
            os.path.dirname(__file__), "../test-data/sce.csv"
        )
        self.marker_yaml_file = os.path.join(
            os.path.dirname(__file__), "../test-data/jackson-2020-markers.yml"
        )

        self.input_expr = pd.read_csv(self.expr_csv_file)
        with open(self.marker_yaml_file, "r") as stream:
            marker_dict = yaml.safe_load(stream)

        self.type_dict = marker_dict["cell_types"]
        self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self._dset = SCDataset(
            include_other_column=True,
            expr_input=self.input_expr,
            marker_dict=self.type_dict,
            design=None,
            device=self._device,
        )

        self.model = CellTypeModel(dset=self._dset, random_seed=42, device=self._device)
        self.model.fit(max_epochs=1)

    def test_basic_instance_creation(self):
        """Testing if the instance is created or not"""
        self.assertIsInstance(self.model, CellTypeModel)

    def test_dtype32(self):
        ds = SCDataset(
            include_other_column=True,
            expr_input=self.input_expr,
            marker_dict=self.type_dict,
            design=None,
            dtype=torch.float32,
        )
        m = CellTypeModel(
            dset=ds, random_seed=42, dtype=torch.float32, device=self._device
        )
        m.fit(max_epochs=1)
        data = m.get_data()
        variables = m.get_variables()
        params = list(data.values()) + list(variables.values())
        comp = [ss.dtype == torch.float32 for ss in params]
        self.assertTrue(all(comp))

    def test_dtype64(self):
        ds = SCDataset(
            include_other_column=True,
            expr_input=self.input_expr,
            marker_dict=self.type_dict,
            design=None,
            dtype=torch.float64,
        )
        m = CellTypeModel(
            dset=ds, random_seed=42, dtype=torch.float64, device=self._device
        )
        m.fit(max_epochs=1)
        data = m.get_data()
        variables = m.get_variables()
        params = list(data.values()) + list(variables.values())
        comp = [ss.dtype == torch.float64 for ss in params]
        self.assertTrue(all(comp))

    def test_trainability(self):
        s = [param.requires_grad for param in self.model._recog.parameters()]
        self.assertTrue(all(s))



import os
import unittest

import pandas as pd
import torch
import yaml

from astir.data import SCDataset


class TestSCDataset(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestSCDataset, self).__init__(*args, **kwargs)

        self.expr_csv_file = os.path.join(
            os.path.dirname(__file__), "../test-data/test_data.csv"
        )

        self.marker_yaml_file = os.path.join(
            os.path.dirname(__file__), "../test-data/jackson-2020-markers.yml"
        )

        self.design_file = os.path.join(
            os.path.dirname(__file__), "../test-data/design.csv"
        )

        # Initializing expected values for unittesting
        # self._param_init_expr_pd()
        self.input_expr = pd.read_csv(self.expr_csv_file, index_col=0)
        with open(self.marker_yaml_file, "r") as stream:
            self.marker_dict = yaml.safe_load(stream)

        self.state_markers = self.marker_dict["cell_states"]

        self.marker_genes = list(
            set([l for s in self.state_markers.values() for l in s])
        )

        self.expr = self.input_expr[self.marker_genes]

        self.design = pd.read_csv(self.design_file, index_col=0)

        self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Initializing the actual model
        self.ds = SCDataset(
            include_other_column=False,
            expr_input=self.input_expr,
            marker_dict=self.state_markers,
            design=None,
            dtype=torch.float32,
            device=self._device,
        )

    def _expr_input_tuple(self):
        pass

    def test_basic_instance_creation(self):

        self.assertIsInstance(self.ds, SCDataset)

    def test_marker_genes(self):
        """Testing if _m_proteins field is declared correctly
        Also tests get_features() and get_n_features() methods
        """
        expected_gene_count = len(self.marker_genes)
        actual_gene_count = self.ds.get_n_features()

        expected_gene_names = sorted(self.marker_genes)
        actual_gene_names = sorted(self.ds.get_features())

        self.assertEqual(expected_gene_count, actual_gene_count)

        self.assertEqual(expected_gene_names, actual_gene_names)

    def test_len_constant_N(self):
        self.assertEqual(self.expr.shape[0], len(self.ds))

    def test_get_classes(self):
        """Testing if _classes field is declared correctly
        Also tests get_classes() and get_n_classes() methods
        """
        expected_class_count = len(self.state_markers.keys())
        actual_class_count = self.ds.get_n_classes()

        expected_classes = sorted(self.state_markers.keys())
        actual_classes = sorted(self.ds.get_classes())

        self.assertEqual(expected_class_count, actual_class_count)

        self.assertEqual(expected_classes, actual_classes)

    def test_marker_mat_not_include_other(self):
        """
        Also tests constant G and C
        """
        expected_G = len(self.marker_genes)
        # actual_G = self.ds.get_protein_amount()

        expected_C = len(self.state_markers)
        # actual_C = self.ds.get_class_amount()

        expected_marker_mat = torch.zeros((expected_G, expected_C)).to(self._device)
        actual_marker_mat = self.ds.get_marker_mat()

        for g, protein in enumerate(sorted(self.marker_genes)):
            for c, state in enumerate(self.state_markers):
                if protein in self.state_markers[state]:
                    expected_marker_mat[g, c] = 1.0

        self.assertTrue(
            torch.all(torch.eq(expected_marker_mat, actual_marker_mat)).item()
        )

    def test_cell_names(self):

        expected_cell_names = sorted(self.expr.index)
        actual_cell_names = sorted(self.ds.get_cell_names())

        self.assertTrue(expected_cell_names, actual_cell_names)

    # To implement: significant
    def test_marker_mat_include_other(self):
        self.type_markers = self.marker_dict["cell_types"]
        self.marker_genes = list(
            set([l for s in self.type_markers.values() for l in s])
        )

        self.ds = SCDataset(
            include_other_column=True,
            expr_input=self.input_expr,
            marker_dict=self.type_markers,
            design=None,
            dtype=torch.float32,
            device=self._device,
        )

        G = self.ds.get_n_features()
        C = self.ds.get_n_classes()

        expected_marker_mat = torch.zeros((G, C + 1)).to(self._device)
        actual_marker_mat = self.ds.get_marker_mat()
        for g, feature in enumerate(sorted(self.marker_genes)):
            for c, cell_class in enumerate(self.type_markers):
                if feature in self.type_markers[cell_class]:
                    expected_marker_mat[g, c] = 1.0

        self.assertTrue(
            torch.all(
                torch.eq(expected_marker_mat, actual_marker_mat).to(self._device)
            ).item()
        )

    def test_fix_design_none(self):

        expected_design = torch.ones((len(self.ds), 1)).to(
            device=self._device, dtype=torch.float64
        )
        actual_design = self.ds.get_design()

        self.assertTrue(
            torch.all(torch.eq(expected_design, actual_design).to(self._device)).item()
        )

    def test_fix_design_not_none(self):
        self.design = self.design.to_numpy()

        self.ds = SCDataset(
            include_other_column=False,
            expr_input=self.input_expr,
            marker_dict=self.state_markers,
            design=self.design,
            dtype=torch.float64,
            device=self._device,
        )

        expected_design = torch.from_numpy(self.design).to(
            device=self._device, dtype=torch.float64
        )
        actual_design = self.ds.get_design()

        self.assertTrue(
            torch.all(torch.eq(expected_design, actual_design).to(self._device)).item()
        )

    def test_dtype(self):
        comp = []
        comp.append(self.ds.get_exprs().dtype == torch.float32)
        comp.append(self.ds.get_design().dtype == torch.float32)
        comp.append(self.ds.get_mu().dtype == torch.float32)
        comp.append(self.ds.get_sigma().dtype == torch.float32)
        self.assertTrue(all(comp))


if __name__ == "__main__":
    unittest.main()



import os
import warnings
from typing import Any

import anndata
import loompy
import matplotlib.cbook
import numpy as np
import pandas as pd
import torch
import yaml
from sklearn.preprocessing import OneHotEncoder

warnings.simplefilter(action="ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=matplotlib.cbook.mplDeprecation)


def from_csv_yaml(
    csv_input: str,
    marker_yaml: str,
    design_csv: str = None,
    random_seed: int = 1234,
    dtype: torch.dtype = torch.float64,
) -> Any:
    """Create an Astir object from an expression CSV and marker YAML

    :param csv_input: Path to input csv containing expression for cells (rows) by proteins (columns). First column is
        cell identifier, and additional column names are gene identifiers.
    :param marker_yaml: Path to input YAML file containing marker gene information. Should include cell_type and cell_state
        entries. See documention.
    :param design_csv: Path to design matrix as a CSV. Rows should be cells, and columns covariates. First column is cell
        identifier, and additional column names are covariate identifiers.
    :param random_seed: The random seed to be used to initialize variables,
        defaults to 1234
    :param dtype: datatype of the model parameters, defaults to torch.float64
    """
    df_gex = pd.read_csv(csv_input, index_col=0)

    design = None
    if design_csv is not None:
        design = pd.read_csv(design_csv, index_col=0)
    with open(marker_yaml, "r") as stream:
        marker_dict = yaml.safe_load(stream)
    from astir.astir import Astir

    return Astir(df_gex, marker_dict, design, random_seed, dtype=dtype)


def from_csv_dir_yaml(
    input_dir: str,
    marker_yaml: str,
    random_seed: int = 1234,
    dtype: torch.dtype = torch.float64,
) -> Any:
    """Create an Astir object a directory containing multiple csv files

    :param input_dir: Path to a directory containing multiple CSV files, each in the format expected by
        `from_csv_yaml`
    :param marker_yaml: Path to input YAML file containing marker gene information. Should include cell_type and cell_state
        entries. See documention.
    :param design_csv: Path to design matrix as a CSV. Rows should be cells, and columns covariates. First column is cell
        identifier, and additional column names are covariate identifiers
    :param random_seed: The random seed to be used to initialize variables,
        defaults to 1234
    :param dtype: datatype of the model parameters, defaults to torch.float64
    """
    # TODO: add text explaining concatenation
    # Parse the input directory
    csv_files = [
        os.path.join(input_dir, f) for f in os.listdir(input_dir) if f.endswith("csv")
    ]

    # Read to gene expression df and parse
    dfs = [pd.read_csv(f, index_col=0) for f in csv_files]
    df_gex = pd.concat(dfs, axis=0)

    # Construct a sample specific design matrix
    design_list = [np.repeat(str(i), dfs[i].shape[0]) for i in range(len(dfs))]
    design = (
        OneHotEncoder()
        .fit_transform(np.concatenate(design_list, axis=0).reshape(-1, 1))
        .todense()
    )
    design = design[:, :-1]  # remove final column
    design = np.concatenate(
        [np.ones((design.shape[0], 1)), design], axis=1
    )  # add in intercept!

    with open(marker_yaml, "r") as stream:
        marker_dict = yaml.safe_load(stream)
    from astir.astir import Astir

    return Astir(df_gex, marker_dict, design, random_seed, dtype)


def from_loompy_yaml(
    loom_file: str,
    marker_yaml: str,
    protein_name_attr: str = "protein",
    cell_name_attr: str = "cell_name",
    batch_name_attr: str = "batch",
    random_seed: int = 1234,
    dtype: torch.dtype = torch.float64,
) -> Any:
    """Create an Astir object from a loom file and a marker yaml

    :param loom_file: Path to a loom file, where rows correspond to proteins and columns to cells
    :param marker_yaml: Path to input YAML file containing marker gene information. Should include cell_type and cell_state
        entries. See documention.
    :param protein_name_attr: The attribute (key) in the row attributes that identifies the protein names
        (required to match with the marker gene information), defaults to
        protein
    :param cell_name_attr: The attribute (key) in the column attributes that
        identifies the name of each cell, defaults to cell_name
    :param batch_name_attr: The attribute (key) in the column attributes that identifies the batch. A design matrix
        will be built using this (if present) using a one-hot encoding to
        control for batch, defaults to batch
    :param random_seed: The random seed to be used to initialize variables,
        defaults to 1234
    :param dtype: datatype of the model parameters, defaults to torch.float64
    :return: An object of class `astir_bash.py.Astir` using data imported from the loom files
    """
    # TODO: This function is memory inefficient and goes against the philosophy of loom files. Should be improved
    batch_list = None
    with loompy.connect(loom_file) as ds:
        df_gex = pd.DataFrame(ds[:, :].T)
        df_gex.columns = ds.ra[protein_name_attr]

        if cell_name_attr in ds.ca.keys():
            df_gex.index = ds.ca[cell_name_attr]

        if batch_name_attr in ds.ca.keys():
            batch_list = ds.ca[batch_name_attr]

    design = None

    if batch_list is not None:
        design = OneHotEncoder().fit_transform(batch_list.reshape(-1, 1)).todense()
        design = design[:, :-1]  # remove final column
        design = np.concatenate([np.ones((design.shape[0], 1)), design], axis=1)

    with open(marker_yaml, "r") as stream:
        marker_dict = yaml.safe_load(stream)
    from astir.astir import Astir

    return Astir(df_gex, marker_dict, design, random_seed, dtype)


def from_anndata_yaml(
    anndata_file: str,
    marker_yaml: str,
    protein_name: str = None,
    cell_name: str = None,
    batch_name: str = "batch",
    random_seed: int = 1234,
    dtype: torch.dtype = torch.float64,
) -> Any:
    """Create an Astir object from an :class:`anndata.Anndata` file and a
        marker yaml

    :param anndata_file: Path to an :class:`anndata.Anndata` `h5py` file
    :param marker_yaml: Path to input YAML file containing marker gene information. Should include cell_type and cell_state
        entries. See documention.
    :param protein_name: The column of `adata.var` containing protein names. If this is none, defaults to `adata.var_names`
    :param cell_name:  The column of `adata.obs` containing cell names. If this is none, defaults to `adata.obs_names`
    :param batch_name: The column of `adata.obs` containing batch names. A design matrix
        will be built using this (if present) using a one-hot encoding to
        control for batch, defaults to 'batch'
    :param random_seed: The random seed to be used to initialize variables,
        defaults to 1234
    :param dtype: datatype of the model parameters, defaults to torch.float64
    :return: An object of class `astir_bash.py.Astir` using data imported from the loom files
    """
    # TODO: This function is memory inefficient and goes against the philosophy of loom files. Should be improved
    batch_list = None

    ad = anndata.read_h5ad(anndata_file)

    df_gex = pd.DataFrame(ad.X)

    if protein_name is not None:
        df_gex.columns = ad.var[protein_name]
    else:
        df_gex.columns = ad.var_names

    if cell_name is not None:
        df_gex.index = ad.obs[cell_name]
    else:
        df_gex.index = ad.obs_names

    if batch_name is not None:
        batch_list = ad.obs[batch_name]

    design = None

    if batch_list is not None:
        design = (
            OneHotEncoder()
            .fit_transform(batch_list.to_numpy().reshape(-1, 1))
            .todense()
        )
        design = design[:, :-1]  # remove final column
        design = np.concatenate([np.ones((design.shape[0], 1)), design], axis=1)

    with open(marker_yaml, "r") as stream:
        marker_dict = yaml.safe_load(stream)
    from astir.astir import Astir

    return Astir(df_gex, marker_dict, design, random_seed, dtype)
